# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Optional, Union

import dace
from dace import properties as dace_properties, transformation as dace_transformation
from dace.sdfg import nodes as dace_nodes
from dace.transformation import helpers as dace_helpers
from dace.transformation.passes import analysis as dace_analysis


@dace_properties.make_properties
class RemoveScalarCopies(dace_transformation.SingleStateTransformation):
    r"""Removes copies between two scalar transient variables.
    Example:
     ___
    /   \
    | A |
    \___/
      |
     \/
     ___
    /   \
    | B |
    \___/
    is transformed to
     ___
    /   \
    | A |
    \___/
    and all uses of B are replaced with A.
    """

    first_access_node = dace_transformation.PatternNode(dace_nodes.AccessNode)
    second_access_node = dace_transformation.PatternNode(dace_nodes.AccessNode)

    assume_single_use_data = dace_properties.Property(
        dtype=bool,
        default=False,
        desc="Always assume that `self.access_node` is single use data. Only useful if used through `SplitAccessNode.apply_to()`.",
    )

    # Name of all data that is used at only one place. Is computed by the
    #  `FindSingleUseData` pass and be passed at construction time. Needed until
    #  [issue#1911](https://github.com/spcl/dace/issues/1911) has been solved.
    _single_use_data: Optional[dict[dace.SDFG, set[str]]]

    def __init__(
        self,
        *args: Any,
        single_use_data: Optional[dict[dace.SDFG, set[str]]] = None,
        assume_single_use_data: Optional[bool] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._single_use_data = single_use_data
        if assume_single_use_data is not None:
            self.assume_single_use_data = assume_single_use_data

    @classmethod
    def expressions(cls) -> Any:
        return [dace.sdfg.utils.node_path_graph(cls.first_access_node, cls.second_access_node)]

    def can_be_applied(
        self,
        graph: Union[dace.SDFGState, dace.SDFG],
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        first_node: dace_nodes.AccessNode = self.first_access_node
        first_node_desc = first_node.desc(sdfg)
        second_node: dace_nodes.AccessNode = self.second_access_node
        second_node_desc = second_node.desc(sdfg)

        scope_dict = graph.scope_dict()

        # Make sure that both access nodes are in the same scope
        if scope_dict[first_node] != scope_dict[second_node]:
            return False

        # Make sure that both access nodes are transients
        if not (first_node_desc.transient and second_node_desc.transient):
            return False

        edges = list(graph.edges_between(first_node, second_node))
        if len(edges) != 1:
            return False
        edge = edges[0]

        # Check that the second access node has only one incoming edge, which is the one from the first access node.
        if graph.in_degree(second_node) != 1:
            return False

        # Check if the edge transfers only one element
        if edge.data.num_elements() != 1:
            return False
        if edge.data.dynamic:
            return False

        # Check that all the outgoing edges of the second access node transfer only one element
        for out_edges in graph.out_edges(second_node):
            if out_edges.data.num_elements() != 1:
                return False

        # Make sure that the data descriptors of both access nodes are scalars
        # TODO(iomaganaris): We could extend this transfromation to handle AccessNodes that are arrays with 1 element as well
        if not isinstance(first_node_desc, dace.data.Scalar) or not isinstance(
            second_node_desc, dace.data.Scalar
        ):
            return False

        # Make sure that both access nodes are not views
        if isinstance(first_node_desc, dace.data.View) or isinstance(
            second_node_desc, dace.data.View
        ):
            return False

        # Make sure that the second AccessNode data is single use data since we're only going to remove that one
        if self.assume_single_use_data:
            single_use_data = {sdfg: {second_node.data}}
        if self._single_use_data is None:
            find_single_use_data = dace_analysis.FindSingleUseData()
            single_use_data = find_single_use_data.apply_pass(sdfg, None)
        else:
            single_use_data = self._single_use_data
        if second_node.data not in single_use_data[sdfg]:
            return False

        return True

    def apply(
        self,
        graph: Union[dace.SDFGState, dace.SDFG],
        sdfg: dace.SDFG,
    ) -> None:
        first_node: dace_nodes.AccessNode = self.first_access_node
        second_node: dace_nodes.AccessNode = self.second_access_node

        # Redirect all outcoming edges of the second access node to the first
        for edge in list(graph.out_edges(second_node)):
            dace_helpers.redirect_edge(
                state=graph,
                edge=edge,
                new_src=first_node,
                new_data=first_node.data if edge.data.data == second_node.data else edge.data.data,
            )

        # Remove the second access node
        graph.remove_node(second_node)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
import functools
import types
from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable, Optional, Union

import devtools

from gt4py._core import definitions as core_defs, types as core_types
from gt4py.next import common, config
from gt4py.next.iterator import builtins, dispatcher
from gt4py.next.iterator.builtins import BackendNotSelectedError, builtin_dispatch
from gt4py.next.otf import arguments
from gt4py.next.program_processors import program_formatter
from gt4py.next.type_system import type_specifications as ts, type_translation as tt


if TYPE_CHECKING:
    # TODO(tehrengruber): remove cirular dependency and import unconditionally
    from gt4py.next import backend as next_backend

__all__ = ["fendef", "fundef", "if_stmt", "offset", "set_at"]


@dataclass(frozen=True)
class Offset:
    value: Union[int, str]


def offset(value):
    return Offset(value)


class CartesianDomain(dict): ...


class UnstructuredDomain(dict): ...


# dependency inversion, register fendef for embedded execution or for tracing/parsing here
# TODO(ricoh): this pattern lead to import cycles with `fendef_codegen`
#   and was changed there. Maybe applies to `fendef_embedded` too?
fendef_embedded: Optional[Callable[[types.FunctionType], None]] = None


@dataclasses.dataclass
class FendefDispatcher:
    definition: types.FunctionType
    offset_provider: Optional[common.OffsetProvider]
    column_axis: Optional[common.Dimension]

    def itir(self, *args):
        # TODO(ricoh): refactor so that `tracing` does not import this module
        #   and can be imported top level. Then set `fendef_tracing` as a
        #   proper default value, instead of using `None` as a sentinel.
        from gt4py.next.iterator.tracing import trace_fencil_definition

        fencil_definition = trace_fencil_definition(self.definition, args)

        if config.DEBUG:
            devtools.debug(fencil_definition)
        return fencil_definition

    def __call__(
        self,
        *args,
        backend: Optional[next_backend.Backend | program_formatter.ProgramFormatter] = None,
        offset_provider=None,
        column_axis=None,
    ):
        offset_provider = offset_provider or self.offset_provider
        column_axis = column_axis or self.column_axis

        if backend is not None:
            itir_node = self.itir(*args)

            # TODO(tehrengruber): remove cirular dependency and place import at the top of the file
            from gt4py.next import backend as next_backend

            if isinstance(backend, next_backend.Backend):
                assert isinstance(backend, next_backend.Backend)
                compiled_program = backend.compile(
                    itir_node,
                    arguments.CompileTimeArgs.from_concrete(
                        *args, offset_provider=offset_provider, column_axis=column_axis
                    ),
                )
                compiled_program(*args, offset_provider=offset_provider)
            elif isinstance(backend, program_formatter.ProgramFormatter):
                return backend(
                    itir_node, *args, offset_provider=offset_provider, column_axis=column_axis
                )
            else:
                raise ValueError(
                    "Backend must be a 'gt4py.next.backend.Backend' or "
                    "'gt4py.next.program_formatter.ProgramFormatter'."
                )
        else:
            if fendef_embedded is None:
                raise RuntimeError("Embedded execution is not registered.")
            fendef_embedded(
                self.definition, *args, offset_provider=offset_provider, column_axis=column_axis
            )


def fendef(
    definition: Optional[types.FunctionType] = None,
    *,
    offset_provider: Optional[common.OffsetProvider] = None,
    column_axis: Optional[common.Dimension] = None,
):
    """
    Dispatches to embedded execution or execution with code generation.

    If `backend` keyword argument is not set or None `fendef_embedded` will be called,
    else `fendef_codegen` will be called.
    """
    if not definition:
        return functools.partial(fendef, offset_provider=offset_provider, column_axis=column_axis)

    return FendefDispatcher(
        definition=definition, offset_provider=offset_provider or {}, column_axis=column_axis
    )


def _deduce_domain(
    domain: dict[common.Dimension, range], offset_provider_type: common.OffsetProviderType
):
    if isinstance(domain, UnstructuredDomain):
        domain_builtin = builtins.unstructured_domain
    elif isinstance(domain, CartesianDomain):
        domain_builtin = builtins.cartesian_domain
    else:
        domain_builtin = (
            builtins.unstructured_domain
            if any(isinstance(o, common.ConnectivityType) for o in offset_provider_type.values())
            else builtins.cartesian_domain
        )

    return domain_builtin(
        *tuple(map(lambda x: builtins.named_range(x[0], x[1].start, x[1].stop), domain.items()))
    )


@dataclass
class FundefFencilWrapper:
    fundef_dispatcher: FundefDispatcher
    domain: Callable | dict

    def _get_fencil(self, offset_provider=None):
        @fendef
        def impl(out, *inps):
            dom = self.domain
            if isinstance(dom, Callable):
                # if domain is expressed as calls to builtin `domain()` we need to pass it lazily
                # as dispatching needs to happen inside of the fencil
                dom = dom()
            elif isinstance(dom, dict):
                # if passed as a dict, we need to convert back to builtins for interpretation by the backends
                assert offset_provider is not None
                dom = _deduce_domain(dom, common.offset_provider_to_type(offset_provider))
            set_at(builtins.as_fieldop(self.fundef_dispatcher, dom)(*inps), dom, out)

        return impl

    def itir(self, *args, **kwargs):
        return self._get_fencil(offset_provider=kwargs.get("offset_provider")).itir(*args, **kwargs)

    def __call__(self, *args, out, **kwargs):
        self._get_fencil(offset_provider=kwargs.get("offset_provider"))(out, *args, **kwargs)

    def format_itir(self, *args, out, **kwargs):
        return self._get_fencil(offset_provider=kwargs.get("offset_provider")).format_itir(
            out, *args, **kwargs
        )


class FundefDispatcher:
    _hook = None
    # hook is an object that
    # - evaluates to true if it should be used,
    # - is callable with an instance of FundefDispatcher
    # - returns callable that takes the function arguments

    def __init__(self, fun) -> None:
        self.fun = fun
        self.__name__ = fun.__name__

    def __getitem__(self, domain):
        return FundefFencilWrapper(self, domain)

    def __call__(self, *args):
        if type(self)._hook:
            return type(self)._hook(self)(*args)
        else:
            return self.fun(*args)

    @classmethod
    def register_hook(cls, hook):
        cls._hook = hook


def fundef(fun):
    return FundefDispatcher(fun)


@builtin_dispatch
def set_at(*args):
    return BackendNotSelectedError()


@builtin_dispatch
def get_domain_range(*args):
    return BackendNotSelectedError()


@builtin_dispatch
def if_stmt(*args):
    return BackendNotSelectedError()


@builtin_dispatch
def temporary(*args):
    return BackendNotSelectedError()


def _dtypebuiltin_to_ts(dtype: callable) -> ts.ScalarType:
    assert isinstance(dtype, dispatcher._fun_dispatcher)
    dtype_name = dtype.fun.__name__
    np_dtype = getattr(core_types, dtype_name)
    return tt.from_dtype(core_defs.dtype(np_dtype))



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
import functools
from typing import Any, Optional, Sequence, cast

import devtools

from gt4py.eve import NodeTranslator, traits
from gt4py.next import common, config, errors, utils
from gt4py.next.ffront import (
    fbuiltins,
    gtcallable,
    program_ast as past,
    stages as ffront_stages,
    transform_utils,
    type_info as ffront_ti,
    type_specifications as ts_ffront,
)
from gt4py.next.ffront.stages import ConcretePASTProgramDef
from gt4py.next.iterator import ir as itir
from gt4py.next.iterator.ir_utils import ir_makers as im
from gt4py.next.iterator.transforms import remap_symbols, replace_get_domain_range_with_constants
from gt4py.next.otf import arguments, definitions, workflow
from gt4py.next.type_system import type_info, type_specifications as ts


# FIXME[#1582](tehrengruber): This should only depend on the program not the arguments. Remove
#  dependency as soon as column axis can be deduced from ITIR in consumers of the CompilableProgram.
def past_to_gtir(inp: ConcretePASTProgramDef) -> definitions.CompilableProgramDef:
    """
    Lower a PAST program definition to Iterator IR.

    Example:
        >>> from gt4py import next as gtx
        >>> from gt4py.next.otf import arguments, toolchain
        >>> IDim = gtx.Dimension("I")

        >>> @gtx.field_operator
        ... def copy(a: gtx.Field[[IDim], gtx.float32]) -> gtx.Field[[IDim], gtx.float32]:
        ...     return a

        >>> @gtx.program
        ... def copy_program(
        ...     a: gtx.Field[[IDim], gtx.float32], out: gtx.Field[[IDim], gtx.float32]
        ... ):
        ...     copy(a, out=out)

        >>> compile_time_args = arguments.CompileTimeArgs(
        ...     args=tuple(param.type for param in copy_program.past_stage.past_node.params),
        ...     kwargs={},
        ...     offset_provider={"I": IDim},
        ...     column_axis=None,
        ...     argument_descriptor_contexts={},
        ... )

        >>> itir_copy = past_to_gtir(
        ...     toolchain.ConcreteArtifact(copy_program.past_stage, compile_time_args)
        ... )

        >>> print(itir_copy.data.id)
        copy_program

        >>> print(type(itir_copy.data))
        <class 'gt4py.next.iterator.ir.Program'>
    """
    all_closure_vars = transform_utils._get_closure_vars_recursively(inp.data.closure_vars)
    offsets_and_dimensions = transform_utils._filter_closure_vars_by_type(
        all_closure_vars, fbuiltins.FieldOffset, common.Dimension
    )
    grid_type = transform_utils._deduce_grid_type(
        inp.data.grid_type, offsets_and_dimensions.values()
    )

    gt_callables = transform_utils._filter_closure_vars_by_type(
        all_closure_vars, gtcallable.GTCallable
    ).values()

    # FIXME[#1582](tehrengruber): remove after refactoring to GTIR
    # TODO(ricoh): The following calls to .__gt_itir__, which will use whatever
    #  backend is set for each of these field operators (GTCallables). Instead
    #  we should use the current toolchain to lower these to ITIR. This will require
    #  making this step aware of the toolchain it is called by (it can be part of multiple).
    lowered_funcs = []
    for gt_callable in gt_callables:
        lowered_funcs.append(gt_callable.__gt_gtir__())

    itir_program = ProgramLowering.apply(
        inp.data.past_node, function_definitions=lowered_funcs, grid_type=grid_type
    )

    # TODO(tehrengruber): Put this in a dedicated transformation step.
    if arguments.StaticArg in inp.args.argument_descriptor_contexts:
        static_arg_descriptors = inp.args.argument_descriptor_contexts[arguments.StaticArg]
        if not all(
            isinstance(arg_descriptor, arguments.StaticArg)
            or all(el is None for el in utils.flatten_nested_tuple(arg_descriptor))  # type: ignore[arg-type]
            for arg_descriptor in static_arg_descriptors.values()
        ):
            raise NotImplementedError("Only top-level arguments can be static.")
        static_args = {
            name: im.literal_from_tuple_value(descr.value)  # type: ignore[union-attr]  # type checked above
            for name, descr in static_arg_descriptors.items()
            if not any(el is None for el in utils.flatten_nested_tuple(descr))  # type: ignore[arg-type]
        }
        body = remap_symbols.RemapSymbolRefs().visit(itir_program.body, symbol_map=static_args)
        itir_program = itir.Program(
            id=itir_program.id,
            function_definitions=itir_program.function_definitions,
            params=itir_program.params,
            declarations=itir_program.declarations,
            body=body,
        )

    # TODO(tehrengruber): Put this in a dedicated transformation step.
    if context := inp.args.argument_descriptor_contexts.get(arguments.FieldDomainDescriptor, None):
        field_domains = {
            param: utils.tree_map(lambda x: x.domain if x is not None else x)(v)
            for param, v in context.items()
        }
        itir_program = (
            replace_get_domain_range_with_constants.ReplaceGetDomainRangeWithConstants.apply(
                itir_program, sizes=field_domains
            )
        )

    # Translate NamedCollectionTypes to TupleTypes in compile-time args
    args = tuple(ffront_ti.named_collections_to_tuple_types(arg) for arg in inp.args.args)
    kwargs: dict[str, ts.TypeSpec] = {
        k: ffront_ti.named_collections_to_tuple_types(v) for k, v in inp.args.kwargs.items()
    }
    compile_time_args = dataclasses.replace(
        inp.args, args=args, kwargs=kwargs, column_axis=_column_axis(all_closure_vars)
    )

    if config.DEBUG or inp.data.debug:
        devtools.debug(itir_program)

    return definitions.CompilableProgramDef(data=itir_program, args=compile_time_args)


def past_to_gtir_factory(
    cached: bool = True,
) -> workflow.Workflow[ConcretePASTProgramDef, definitions.CompilableProgramDef]:
    wf = workflow.make_step(past_to_gtir)
    if cached:
        wf = workflow.CachedStep(wf, hash_function=ffront_stages.fingerprint_stage)
    return wf


def _column_axis(all_closure_vars: dict[str, Any]) -> Optional[common.Dimension]:
    # construct mapping from column axis to scan operators defined on
    #  that dimension. only one column axis is allowed, but we can use
    #  this mapping to provide good error messages.
    scanops_per_axis: dict[common.Dimension, list[str]] = {}
    for name, gt_callable in transform_utils._filter_closure_vars_by_type(
        all_closure_vars, gtcallable.GTCallable
    ).items():
        if isinstance((type_ := gt_callable.__gt_type__()), ts_ffront.ScanOperatorType):
            scanops_per_axis.setdefault(type_.axis, []).append(name)

    if len(scanops_per_axis.values()) == 0:
        return None

    if len(scanops_per_axis.values()) != 1:
        scanops_per_axis_str = "\n".join(
            f"- {dim.value}: {', '.join(scanops)}" for dim, scanops in scanops_per_axis.items()
        )

        raise TypeError(
            "Only 'ScanOperator's defined on the same axis "
            f"can be used in a 'Program', found:\n{scanops_per_axis_str}\n"
        )

    return iter(scanops_per_axis.keys()).__next__()


def _compute_field_slice(node: past.Subscript) -> list[past.Slice]:
    out_field_name: past.Name = node.value
    out_field_slice_: list[past.Slice]
    if isinstance(node.slice_, past.TupleExpr) and all(
        isinstance(el, past.Slice) for el in node.slice_.elts
    ):
        out_field_slice_ = cast(list[past.Slice], node.slice_.elts)  # type ensured by if
    elif isinstance(node.slice_, past.Slice):
        out_field_slice_ = [node.slice_]
    else:
        raise AssertionError(
            "Unexpected 'out' argument, must be tuple of slices or slice expression."
        )
    node_dims = cast(ts.FieldType, node.type).dims
    assert isinstance(node_dims, list)
    if isinstance(node.type, ts.FieldType) and len(out_field_slice_) != len(node_dims):
        raise errors.DSLError(
            node.location,
            f"Too many indices for field '{out_field_name}': field is {len(node_dims)}"
            f"-dimensional, but {len(out_field_slice_)} were indexed.",
        )
    return out_field_slice_


def _get_element_from_tuple_expr(node: past.Expr, path: tuple[int, ...]) -> past.Expr:
    """Get element from a (nested) TupleExpr by following the given path.

    Pre-condition: `node` is a `past.TupleExpr` (if `path ! = ()`)
    and `path` is a valid path through the nested tuple structure.
    """
    return functools.reduce(lambda e, i: e.elts[i], path, node)  # type: ignore[attr-defined] # see pre-condition


def _unwrap_tuple_expr(expr: past.Expr, path: tuple[int, ...]) -> tuple[past.Expr, Sequence[int]]:
    """Unwrap (nested) TupleExpr by following the given path as long as possible.

    If a non-tuple expression is encountered, the current expression and the remaining path are
    returned.
    """
    path_remainder: Sequence[int] = path
    while isinstance(expr, past.TupleExpr):
        idx, *path_remainder = path_remainder
        expr = expr.elts[idx]

    return expr, path_remainder


@dataclasses.dataclass
class ProgramLowering(
    traits.PreserveLocationVisitor, traits.VisitorWithSymbolTableTrait, NodeTranslator
):
    """
    Lower Program AST (PAST) to Iterator IR (ITIR).

    Examples
    --------
    >>> from gt4py.next.ffront.func_to_past import ProgramParser
    >>> from gt4py.next.iterator import ir
    >>> from gt4py.next import Dimension, Field
    >>>
    >>> float64 = float
    >>> IDim = Dimension("IDim")
    >>>
    >>> def fieldop(inp: Field[[IDim], "float64"]) -> Field[[IDim], "float64"]: ...
    >>> def program(inp: Field[[IDim], "float64"], out: Field[[IDim], "float64"]):
    ...     fieldop(inp, out=out)
    >>>
    >>> parsed = ProgramParser.apply_to_function(program)  # doctest: +SKIP
    >>> fieldop_def = ir.FunctionDefinition(
    ...     id="fieldop",
    ...     params=[ir.Sym(id="inp")],
    ...     expr=ir.FunCall(fun=ir.SymRef(id="deref"), pos_only_args=[ir.SymRef(id="inp")]),
    ... )  # doctest: +SKIP
    >>> lowered = ProgramLowering.apply(
    ...     parsed, [fieldop_def], grid_type=common.GridType.CARTESIAN
    ... )  # doctest: +SKIP
    >>> type(lowered)  # doctest: +SKIP
    <class 'gt4py.next.iterator.ir.Program'>
    >>> lowered.id  # doctest: +SKIP
    SymbolName('program')
    >>> lowered.params  # doctest: +SKIP
    [Sym(id=SymbolName('inp')), Sym(id=SymbolName('out')), Sym(id=SymbolName('__inp_size_0')), Sym(id=SymbolName('__out_size_0'))]
    """

    grid_type: common.GridType

    # TODO(tehrengruber): enable doctests again. For unknown / obscure reasons
    #  the above doctest fails when executed using `pytest --doctest-modules`.

    @classmethod
    def apply(
        cls,
        node: past.Program,
        function_definitions: list[itir.FunctionDefinition],
        grid_type: common.GridType,
    ) -> itir.Program:
        return cls(grid_type=grid_type).visit(node, function_definitions=function_definitions)

    def visit_Program(
        self,
        node: past.Program,
        *,
        function_definitions: list[itir.FunctionDefinition],
        **kwargs: Any,
    ) -> itir.Program:
        # The ITIR does not support dynamically getting the size of a field. As
        #  a workaround we add additional arguments to the fencil definition
        #  containing the size of all fields. The caller of a program is (e.g.
        #  program decorator) is required to pass these arguments.

        params = self.visit(node.params)

        set_ats = [self._visit_field_operator_call(stmt, **kwargs) for stmt in node.body]
        return itir.Program(
            id=node.id,
            function_definitions=function_definitions,
            params=params,
            declarations=[],
            body=set_ats,
        )

    def _visit_field_operator_call(self, node: past.Call, **kwargs: Any) -> itir.SetAt:
        assert isinstance(node.kwargs["out"].type, ts.TypeSpec)
        assert type_info.is_type_or_tuple_of_type(node.kwargs["out"].type, ts.FieldType)

        node_kwargs = {**node.kwargs}
        domain = node_kwargs.pop("domain", None)
        output, lowered_domain = self._visit_stencil_call_out_arg(
            node_kwargs.pop("out"), domain, **kwargs
        )

        assert isinstance(node.func.type, (ts_ffront.FieldOperatorType, ts_ffront.ScanOperatorType))

        args, node_kwargs = type_info.canonicalize_arguments(node.func.type, node.args, node_kwargs)

        lowered_args, lowered_kwargs = self.visit(args, **kwargs), self.visit(node_kwargs, **kwargs)

        return itir.SetAt(
            expr=im.call(node.func.id)(*lowered_args, *lowered_kwargs.values()),
            domain=lowered_domain,
            target=output,
        )

    def _visit_slice_bound(
        self,
        slice_bound: Optional[past.Constant],
        default_value: itir.Expr,
        start_idx: itir.Expr,
        stop_idx: itir.Expr,
        **kwargs: Any,
    ) -> itir.Expr:
        if slice_bound is None:
            lowered_bound = default_value
        elif isinstance(slice_bound, past.Constant):
            assert isinstance(slice_bound.type, ts.ScalarType) and type_info.is_integral(
                slice_bound.type
            )
            if slice_bound.value < 0:
                lowered_bound = im.plus(stop_idx, self.visit(slice_bound, **kwargs))
            else:
                lowered_bound = im.plus(start_idx, self.visit(slice_bound, **kwargs))
        else:
            raise AssertionError("Expected 'None' or 'past.Constant'.")
        if slice_bound:
            lowered_bound.location = slice_bound.location
        return lowered_bound

    def _construct_itir_out_arg(self, node: past.Expr) -> itir.Expr:
        if isinstance(node, past.Name):
            return itir.SymRef(id=node.id, location=node.location)
        elif isinstance(node, past.Subscript):
            itir_node = self._construct_itir_out_arg(node.value)
            itir_node.location = node.location
            return itir_node
        elif isinstance(node, past.TupleExpr):
            return itir.FunCall(
                fun=itir.SymRef(id="make_tuple"),
                args=[self._construct_itir_out_arg(el) for el in node.elts],
                location=node.location,
            )
        else:
            raise ValueError(
                "Unexpected 'out' argument. Must be a 'past.Name', 'past.Subscript'"
                " or a 'past.TupleExpr' thereof."
            )

    def _construct_itir_domain_arg(
        self,
        out_expr: itir.Expr,
        out_type: ts.FieldType,
        node_domain: Optional[past.Expr],
        slices: Optional[list[past.Slice]] = None,
    ) -> itir.FunCall:
        domain_args = []
        for dim_i, dim in enumerate(out_type.dims):
            # an expression for the range of a dimension
            dim_range = im.call("get_domain_range")(
                out_expr, itir.AxisLiteral(value=dim.value, kind=dim.kind)
            )

            dim_start, dim_stop = im.tuple_get(0, dim_range), im.tuple_get(1, dim_range)
            # bounds
            lower: itir.Expr
            upper: itir.Expr
            if node_domain is not None:
                assert isinstance(node_domain, past.Dict)
                lower, upper = self._construct_itir_initialized_domain_arg(dim_i, dim, node_domain)
            else:
                lower = self._visit_slice_bound(
                    slices[dim_i].lower if slices else None,
                    dim_start,
                    dim_start,
                    dim_stop,
                )
                upper = self._visit_slice_bound(
                    slices[dim_i].upper if slices else None,
                    dim_stop,
                    dim_start,
                    dim_stop,
                )

            if dim.kind == common.DimensionKind.LOCAL:
                raise ValueError(f"common.Dimension '{dim.value}' must not be local.")
            domain_args.append(
                itir.FunCall(
                    fun=itir.SymRef(id="named_range"),
                    args=[itir.AxisLiteral(value=dim.value, kind=dim.kind), lower, upper],
                )
            )

        if self.grid_type == common.GridType.CARTESIAN:
            domain_builtin = "cartesian_domain"
        elif self.grid_type == common.GridType.UNSTRUCTURED:
            domain_builtin = "unstructured_domain"
        else:
            raise AssertionError()

        return itir.FunCall(
            fun=itir.SymRef(id=domain_builtin),
            args=domain_args,
            location=(node_domain or out_expr).location,
        )

    def _construct_itir_initialized_domain_arg(
        self, dim_i: int, dim: common.Dimension, node_domain: past.Dict
    ) -> list[itir.FunCall]:
        assert len(node_domain.values_[dim_i].elts) == 2
        keys_dims_types = cast(ts.DimensionType, node_domain.keys_[dim_i].type).dim
        if keys_dims_types != dim:
            raise ValueError(
                "common.Dimensions in out field and field domain are not equivalent:"
                f"expected '{dim}', got '{keys_dims_types}'."
            )

        return [self.visit(bound) for bound in node_domain.values_[dim_i].elts]

    def _split_field_and_slice(
        self, field: past.Name | past.Subscript
    ) -> tuple[past.Name, list[past.Slice] | None]:
        if isinstance(field, past.Subscript):
            return field.value, _compute_field_slice(field)
        else:
            assert isinstance(field, past.Name)
            return field, None

    def _visit_stencil_call_out_arg(
        self, out_arg: past.Expr, domain_arg: Optional[past.Expr], **kwargs: Any
    ) -> tuple[itir.Expr, itir.FunCall]:
        assert isinstance(out_arg, (past.Subscript, past.Name, past.TupleExpr)), (
            "Unexpected 'out' argument. Must be a 'past.Subscript', 'past.Name' or 'past.TupleExpr' node."
        )

        @utils.tree_map(
            collection_type=ts.COLLECTION_TYPE_SPECS,
            with_path_arg=True,
            unpack=True,
            result_collection_constructor=lambda _, elts: im.make_tuple(*elts),
        )
        def impl(out_type: ts.FieldType, path: tuple[int, ...]) -> tuple[itir.Expr, itir.Expr]:
            out_field, path_remainder = _unwrap_tuple_expr(out_arg, path)

            assert isinstance(out_field, (past.Name, past.Subscript))
            out_field, slice_info = self._split_field_and_slice(out_field)

            domain_element = (
                _get_element_from_tuple_expr(domain_arg, path)
                if isinstance(domain_arg, past.TupleExpr)
                else domain_arg
            )

            lowered_out_field = functools.reduce(
                lambda expr, i: im.tuple_get(i, expr), path_remainder, self.visit(out_field)
            )
            lowered_domain = self._construct_itir_domain_arg(
                lowered_out_field,
                out_type,
                domain_element,
                slice_info,
            )
            return lowered_out_field, lowered_domain

        return impl(out_arg.type)

    def visit_Constant(self, node: past.Constant, **kwargs: Any) -> itir.Literal:
        if isinstance(node.type, ts.ScalarType) and node.type.shape is None:
            match node.type.kind:
                case ts.ScalarKind.STRING:
                    raise NotImplementedError(
                        f"Scalars of kind '{node.type.kind}' not supported currently."
                    )
            typename = node.type.kind.name.lower()
            return im.literal(str(node.value), typename)

        raise NotImplementedError("Only scalar literals supported currently.")

    def visit_Name(self, node: past.Name, **kwargs: Any) -> itir.SymRef:
        return itir.SymRef(id=node.id, location=node.location)

    def visit_Symbol(self, node: past.Symbol, **kwargs: Any) -> itir.Sym:
        if isinstance(node.type, ts.COLLECTION_TYPE_SPECS):
            new_symbol_type: ts.TypeSpec = ffront_ti.named_collections_to_tuple_types(node.type)
        else:
            new_symbol_type = node.type
        return itir.Sym(id=node.id, type=new_symbol_type)

    def visit_BinOp(self, node: past.BinOp, **kwargs: Any) -> itir.FunCall:
        return itir.FunCall(
            fun=itir.SymRef(id=node.op.value),
            args=[self.visit(node.left, **kwargs), self.visit(node.right, **kwargs)],
        )

    def visit_Call(self, node: past.Call, **kwargs: Any) -> itir.FunCall:
        if node.func.id in ["maximum", "minimum"]:
            assert len(node.args) == 2
            return itir.FunCall(
                fun=itir.SymRef(id=node.func.id),
                args=[self.visit(node.args[0]), self.visit(node.args[1])],
            )
        else:
            raise NotImplementedError("Only 'minimum', and 'maximum' builtins supported currently.")



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from gt4py.next.ffront.experimental import as_offset, concat_where


__all__ = [
    "as_offset",
    "concat_where",
]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import copy
from typing import Any, Optional, Union

import dace
from dace import (
    data as dace_data,
    properties as dace_properties,
    subsets as dace_subsets,
    transformation as dace_transformation,
)
from dace.sdfg import (
    graph as dace_graph,
    nodes as dace_nodes,
    propagation as dace_propagation,
    utils as dace_sdutils,
)
from dace.transformation import helpers as dace_helpers

from gt4py.next import common as gtx_common
from gt4py.next.program_processors.runners.dace import lowering as gtx_dace_lowering


@dace_properties.make_properties
class LoopBlocking(dace_transformation.SingleStateTransformation):
    """Applies loop blocking, also known as k-blocking, on a Map.

    This transformation takes a multidimensional Map and performs blocking on a
    single dimension, the loop variable is called `I` here. All dimensions except
    `I` are unaffected by this transformation. In the outer Map will be replace the
    `I` range, currently `I = 0:N`, with `__coarse_I = 0:N:B`, where `N` is the
    original size of the range and `B` is the blocking size. The transformation
    will then create an inner sequential map with `I = __coarse_I:(__coarse_I + B)`.

    What makes this transformation different from simple blocking, is that
    the inner map will not just be inserted right after the outer Map.
    Instead the transformation will first identify all nodes that does not depend
    on the blocking parameter `I`, called independent nodes and relocate them
    between the outer and inner map. Note that an independent node must be connected
    to the MapEntry or another independent node.
    Thus these operations will only be performed once, per outer loop iteration.

    Args:
        blocking_size: The size of the block, denoted as `B` above.
        blocking_parameter: On which parameter should we block.
        require_independent_nodes: If `True` only apply loop blocking if the Map
            actually contains independent nodes. Defaults to `True`.

    Todo:
        - Modify the inner map such that it always starts at zero.
        - Allow more than one parameter on which we block.
    """

    blocking_size = dace_properties.Property(
        dtype=int,
        allow_none=True,
        desc="Size of the inner blocks; 'B' in the above description.",
    )
    blocking_parameter = dace_properties.Property(
        dtype=str,
        allow_none=True,
        desc="Name of the iteration variable on which to block (must be an exact match);"
        " 'I' in the above description.",
    )
    require_independent_nodes = dace_properties.Property(
        dtype=bool,
        default=True,
        desc="If 'True' then blocking is only applied if there are independent nodes.",
    )

    # Set of nodes that are independent of the blocking parameter.
    _independent_nodes: Optional[set[dace_nodes.AccessNode]]
    _dependent_nodes: Optional[set[dace_nodes.AccessNode]]

    outer_entry = dace_transformation.PatternNode(dace_nodes.MapEntry)

    def __init__(
        self,
        blocking_size: Optional[int] = None,
        blocking_parameter: Optional[Union[gtx_common.Dimension, str]] = None,
        require_independent_nodes: Optional[bool] = None,
    ) -> None:
        super().__init__()
        if isinstance(blocking_parameter, gtx_common.Dimension):
            blocking_parameter = gtx_dace_lowering.get_map_variable(blocking_parameter)
        if blocking_parameter is not None:
            self.blocking_parameter = blocking_parameter
        if blocking_size is not None:
            self.blocking_size = blocking_size
        if require_independent_nodes is not None:
            self.require_independent_nodes = require_independent_nodes
        self._independent_nodes = None
        self._dependent_nodes = None

    @classmethod
    def expressions(cls) -> Any:
        return [dace.sdfg.utils.node_path_graph(cls.outer_entry)]

    def can_be_applied(
        self,
        graph: Union[dace.SDFGState, dace.SDFG],
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        """Test if the map can be blocked.

        The test involves:
        - The map must be at the top level.
        - The block dimension must be present (exact string match).
        - The map range must have step size of 1.
        - The partition must exists (see `partition_map_output()`).
        """
        if self.blocking_parameter is None:
            raise ValueError("The blocking dimension was not specified.")
        elif self.blocking_size is None:
            raise ValueError("The blocking size was not specified.")

        outer_entry: dace_nodes.MapEntry = self.outer_entry
        map_params: list[str] = outer_entry.map.params
        map_range: dace_subsets.Range = outer_entry.map.range
        block_var: str = self.blocking_parameter

        scope = graph.scope_dict()
        if scope[outer_entry] is not None:
            return False

        if block_var not in map_params:
            return False

        block_var_idx = map_params.index(block_var)
        map_range_size = map_range.size()
        if map_range[block_var_idx][2] != 1:
            return False

        # Require that there are more iteration than the blocking size.
        # TODO(phimuell): Synchronize this with the GPU block size since it also
        #   plays into it.
        if (map_range_size[block_var_idx] <= self.blocking_size) == True:  # noqa: E712 [true-false-comparison]  # SymPy Fancy comparison.
            return False

        if not self.partition_map_output(graph, sdfg):
            return False
        self._independent_nodes = None
        self._dependent_nodes = None

        return True

    def apply(
        self,
        graph: Union[dace.SDFGState, dace.SDFG],
        sdfg: dace.SDFG,
    ) -> None:
        """Creates a blocking map.

        Performs the operation described in the doc string.
        """
        # Now compute the partitions of the nodes.
        self.partition_map_output(graph, sdfg)

        # Modify the outer map and create the inner map.
        (outer_entry, outer_exit), (inner_entry, inner_exit) = self._prepare_inner_outer_maps(graph)

        # Reconnect the edges
        self._rewire_map_scope(
            outer_entry=outer_entry,
            outer_exit=outer_exit,
            inner_entry=inner_entry,
            inner_exit=inner_exit,
            state=graph,
            sdfg=sdfg,
        )
        self._independent_nodes = None
        self._dependent_nodes = None

    def _prepare_inner_outer_maps(
        self,
        state: dace.SDFGState,
    ) -> tuple[
        tuple[dace_nodes.MapEntry, dace_nodes.MapExit],
        tuple[dace_nodes.MapEntry, dace_nodes.MapExit],
    ]:
        """Prepare the maps for the blocking.

        The function modifies the outer map, `self.outer_entry`, by replacing the
        blocking parameter, `self.blocking_parameter`, with a coarsened version
        of it. In addition the function will then create the inner map, that
        iterates over the blocking parameter, and these bounds are determined
        by the coarsened blocking parameter of the outer map.

        Args:
            state: The state on which we operate.

        Return:
            The function returns a tuple of length two, the first element is the
            modified outer map and the second element is the newly created
            inner map. Each element consist of a pair containing the map entry
            and map exit nodes of the corresponding maps.
        """
        outer_entry: dace_nodes.MapEntry = self.outer_entry
        outer_exit: dace_nodes.MapExit = state.exit_node(outer_entry)
        outer_map: dace_nodes.Map = outer_entry.map
        outer_range: dace_subsets.Range = outer_entry.map.range
        outer_params: list[str] = outer_entry.map.params
        blocking_parameter_dim = outer_params.index(self.blocking_parameter)

        # This is the name of the iterator that we use in the outer map for the
        #  blocked dimension
        # NOTE: The way we create the new parameter name is important such that it
        #   could be matched through the `unit_strides_kind` argument, which is
        #   the case with this approach. But it makes the `unit_strides_dim`
        #   argument `gt_set_iteration_order()` inapplicable.
        coarse_block_var = "__gtx_coarse_" + self.blocking_parameter

        # Generate the sequential inner map
        rng_start = outer_range[blocking_parameter_dim][0]
        rng_stop = outer_range[blocking_parameter_dim][1]
        inner_label = f"inner_{outer_map.label}"
        inner_range = {
            self.blocking_parameter: dace_subsets.Range.from_string(
                f"(({rng_start}) + ({coarse_block_var}) * ({self.blocking_size})):"
                f"min(({rng_start}) + ({coarse_block_var} + 1) * ({self.blocking_size}), ({rng_stop}) + 1)"
            )
        }
        inner_entry, inner_exit = state.add_map(
            name=inner_label,
            ndrange=inner_range,
            schedule=dace.dtypes.ScheduleType.Sequential,
            debuginfo=copy.copy(outer_map.debuginfo),
        )

        # TODO(phimuell): Investigate if we want to prevent unrolling here

        # Now we modify the properties of the outer map.
        coarse_block_range = dace_subsets.Range.from_string(
            f"0:int_ceil((({rng_stop}) + 1) - ({rng_start}), ({self.blocking_size}))"
        ).ranges[0]
        outer_map.params[blocking_parameter_dim] = coarse_block_var
        outer_map.range[blocking_parameter_dim] = coarse_block_range

        return ((outer_entry, outer_exit), (inner_entry, inner_exit))

    def partition_map_output(
        self,
        state: dace.SDFGState,
        sdfg: dace.SDFG,
    ) -> bool:
        """Computes the partition the of the nodes of the Map.

        The function divides the nodes into two sets, defined as:
        - The independent nodes `\\mathcal{I}`:
            These are the nodes, whose output does not depend on the blocked
            dimension. These nodes can be relocated between the outer and inner map.
            Nodes in these set does not necessarily have a direct edge to map entry.
            However, the exists a path from `outer_entry` to any node in this set.
        - The dependent nodes `\\mathcal{D}`:
            These are the nodes, whose output depend on the blocked dimension.
            Thus they can not be relocated between the two maps, but will remain
            inside the inner scope. All these nodes have at least one edge to map entry.

        The function uses an iterative method to compute the set of independent nodes.
        In each iteration the function will look classify all nodes that have an
        incoming edge originating either at outer_entry or from a node that was
        already classified as independent. This is repeated until no new independent
        nodes are found. This means that independent nodes does not necessarily have
        a direct connection to map entry.

        The dependent nodes on the other side always have a direct edge to outer_entry.
        As they are the set of nodes that are adjacent to outer_entry but are not
        independent.

        For the sake of arguments, all nodes, except the map entry and exit nodes,
        that are inside the scope and are not classified as dependent or independent
        are known as "inner nodes".

        If the function is able to compute the partition `True` is returned and the
        member variables are updated. If the partition does not exists the function
        will return `False` and the respective member variables will be `None`.

        The function will honor `self.require_independent_nodes`. Thus if no independent
        nodes were found the function behaves as if the partition does not exist.

        Args:
            state: The state on which we operate.
            sdfg: The SDFG in which we operate on.

        Note:
            - The function will only inspect the direct children of the map entry.
            - Currently this function only considers the input Memlets and the
                `used_symbol` properties of a Tasklet to determine if a Tasklet is dependent.
        """

        # Clear the previous partition.
        self._independent_nodes = set()
        self._dependent_nodes = None

        while True:
            # Find all the nodes that we have to classify in this iteration.
            #  - All nodes adjacent to `outer_entry` (which is
            #       independent by definition).
            #  - All nodes adjacent to independent nodes.
            nodes_to_classify: set[dace_nodes.Node] = {
                edge.dst for edge in state.out_edges(self.outer_entry)
            }
            for independent_node in self._independent_nodes:
                nodes_to_classify.update({edge.dst for edge in state.out_edges(independent_node)})
            nodes_to_classify.difference_update(self._independent_nodes)

            # Now classify each node
            found_new_independent_node = False
            for node_to_classify in nodes_to_classify:
                class_res = self._classify_node(
                    node_to_classify=node_to_classify,
                    state=state,
                    sdfg=sdfg,
                )

                # Check if the partition exists.
                if class_res is None:
                    self._independent_nodes = None
                    return False
                if class_res is True:
                    found_new_independent_node = True

            # If we found a new independent node then we have to continue.
            if not found_new_independent_node:
                break

        # Perform some cleaning on the independent nodes.
        self._post_process_independent_nodes(state)

        assert all(
            all(
                iedge.src in self._independent_nodes or iedge.src is self.outer_entry
                for iedge in state.in_edges(inode)
            )
            for inode in self._independent_nodes
        )

        # If requested check if the blocking is a good idea.
        if self.require_independent_nodes and (not self._check_if_blocking_is_favourable(state)):
            self._independent_nodes = None
            return False

        # After the independent set is computed compute the set of dependent nodes
        #  as the set of all nodes adjacent to `outer_entry` that are not independent.
        self._dependent_nodes = {
            edge.dst
            for edge in state.out_edges(self.outer_entry)
            if edge.dst not in self._independent_nodes
        }

        return True

    def _classify_node(
        self,
        node_to_classify: dace_nodes.Node,
        state: dace.SDFGState,
        sdfg: dace.SDFG,
    ) -> bool | None:
        """Internal function for classifying a single node.

        The general rule to classify if a node is independent are:
        - The node must be a Tasklet or an AccessNode, in all other cases the
            partition does not exist.
        - `free_symbols` of the nodes shall not contain the blocking parameter.
        - All incoming _empty_ edges must be connected to the map entry.
        - A node either has only empty Memlets or none of them.
        - Incoming Memlets does not depend on the blocking parameter.
        - All incoming edges must start either at `self.outer_entry` or at dependent nodes.
        - All output Memlets are non empty.

        It is important that to realize that the function will add the node to
        classify to `self._independent_nodes` on its own. It is also important that
        the function might add other nodes beside `node_to_classify`.

        Returns:
            The function returns `True` if `node_to_classify` is considered independent.
            If the function returns `False` the node was classified as a dependent node.
            The function will return `None` if the node can not be classified, in this
            case the partition does not exist.

        Args:
            node_to_classify: The node that should be classified.
            state: The state containing the map.
            sdfg: The SDFG that is processed.
        """
        assert self._independent_nodes is not None  # silence MyPy
        outer_entry: dace_nodes.MapEntry = self.outer_entry  # for caching.
        outer_exit: dace_nodes.MapExit = state.exit_node(outer_entry)

        # The node needs to have an input and output.
        if state.in_degree(node_to_classify) == 0 or state.out_degree(node_to_classify) == 0:
            return None

        # The outer MapExit is always classified as dependent.
        if node_to_classify is outer_exit:
            return False

        # To fully understand what is going on we have to look at the input and output
        #  edges of the node. We define them here to allow to modify them in certain
        #  cases.
        in_edges: list[dace_graph.MultiConnectorEdge[dace.Memlet]] = list(
            state.in_edges(node_to_classify)
        )
        out_edges: list[dace_graph.MultiConnectorEdge[dace.Memlet]] = list(
            state.out_edges(node_to_classify)
        )

        # Despite its type the node's free symbols can not contain the blocking
        #  parameter. In case of a Tasklet this would be the body of the Tasklet.
        if self.blocking_parameter in node_to_classify.free_symbols:
            return False

        # If the test succeed then these are the nodes we additionally consider
        #  as independent.
        new_independent_nodes: set[dace_nodes.Node] = {node_to_classify}

        # We are only able to handle certain kind of nodes, so screening them.
        if isinstance(node_to_classify, dace_nodes.Tasklet):
            if node_to_classify.side_effects:
                return None

            # A Tasklet must write to an AccessNode, because otherwise there would
            #  be nothing that could be used to cache anything. Furthermore, this
            #  AccessNode must be outside of the inner loop, i.e. be independent.
            # TODO: Make this check stronger to ensure that there is always an
            #   AccessNode that is independent.
            if not all(
                isinstance(out_edge.dst, dace_nodes.AccessNode)
                for out_edge in state.out_edges(node_to_classify)
                if not out_edge.data.is_empty()
            ):
                return False

        elif isinstance(node_to_classify, dace.nodes.NestedSDFG):
            # Same check as for Tasklets applies to the outputs of a nested SDFG node
            if not all(
                isinstance(out_edge.dst, dace_nodes.AccessNode)
                for out_edge in state.out_edges(node_to_classify)
                if not out_edge.data.is_empty()
            ):
                return False

            # Additionally, test if the symbol mapping depends on the block variable.
            for v in node_to_classify.symbol_mapping.values():
                if self.blocking_parameter in v.free_symbols:
                    return False

        elif isinstance(node_to_classify, dace_nodes.AccessNode):
            # AccessNodes need to have some special properties.
            node_desc: dace.data.Data = node_to_classify.desc(sdfg)
            if isinstance(node_desc, dace.data.View):
                # Views are forbidden.
                return None

            # The access node inside either has scope lifetime or is a scalar.
            if isinstance(node_desc, dace.data.Scalar):
                pass
            elif node_desc.lifetime != dace.dtypes.AllocationLifetime.Scope:
                return None

        elif isinstance(node_to_classify, dace_nodes.MapEntry):
            # We check a Map as a whole and add it as a whole to the set of the
            #  independent nodes. For that reason we must modify `out_edges` since
            #  now the outputs of the associated `MapExit` must be checked.
            #  However, we have to run some Map specific checks.
            # TODO(phimuell): Do we also have to modify `node_to_classify`?
            map_exit = state.exit_node(node_to_classify)

            # The blocking parameter can not be used inside the Map scope.
            map_scope = state.scope_subgraph(node_to_classify)
            if self.blocking_parameter in map_scope.free_symbols:
                return False

            # There is an obscure case, where the Memlet on the inside of a Map scope
            #  and the Memlet on the outside refer to different data, thus we have to
            #  check that here. Note that we only have to do it here in this case
            #  because normally it would be spotted above where we checked the input.
            out_edges = list(state.out_edges(map_exit))
            if any(self.blocking_parameter in out_edge.data.free_symbols for out_edge in out_edges):
                return False

            # Add all nodes of the Map scope, including entry and exit node to the
            #  set of new independent nodes.
            new_independent_nodes.update(map_scope.nodes())

        elif isinstance(node_to_classify, dace.libraries.standard.nodes.Reduce):
            # The only checks we impose on them is the free symbols check and the
            #  input output checks.
            pass

        else:
            # Any other node type we can not handle, so the partition can not exist.
            # TODO(phimuell): Try to handle certain kind of library nodes.
            return None

        # In a first phase we will only look if the partition exists or not.
        #  We will therefore not check if the node is independent or not, since
        #  for these classification to make sense the partition has to exist in the
        #  first place.

        # There are some very small requirements that we impose on the output edges.
        for out_edge in out_edges:
            # We consider nodes that are directly connected to the outer map exit as
            #  dependent. This is an implementation detail to avoid some hard cases.
            #  The only exceptions are scalars which we will handle later.
            # NOTE: We restrict ourself to scalars to make sure that no dynamic
            #  allocation is needed. It could be extended to arrays later.
            if out_edge.dst is outer_exit:
                if (
                    isinstance(node_to_classify, dace_nodes.AccessNode)
                    and isinstance(node_to_classify.desc(sdfg), dace_data.Scalar)
                    and out_edge.data.wcr is None
                ):
                    continue
                return False

        # Now we have to look at incoming edges individually.
        #  We will inspect the subset of the Memlet to see if they depend on the
        #  block variable. If this loop ends normally, then we classify the node
        #  as independent and the node is added to `_independent_nodes`.
        for in_edge in in_edges:
            memlet: dace.Memlet = in_edge.data
            src_subset: dace_subsets.Subset | None = memlet.src_subset
            dst_subset: dace_subsets.Subset | None = memlet.dst_subset

            if memlet.is_empty():  # Empty Memlets do not impose any restrictions.
                continue

            # Now we have to look at the source and destination set of the Memlet.
            subsets_to_inspect: list[dace_subsets.Subset] = []
            if dst_subset is not None:
                subsets_to_inspect.append(dst_subset)
            if src_subset is not None:
                subsets_to_inspect.append(src_subset)

            # If a subset needs the block variable then the node is not independent
            #  but dependent.
            if any(self.blocking_parameter in subset.free_symbols for subset in subsets_to_inspect):
                return False

            # The edge must either originate from `outer_entry` or from an independent
            #  node if not it is dependent.
            if not (in_edge.src is outer_entry or in_edge.src in self._independent_nodes):
                return False

        # Loop ended normally, thus we update the list of independent nodes.
        self._independent_nodes.update(new_independent_nodes)
        return True

    def _post_process_independent_nodes(
        self,
        state: dace.SDFGState,
    ) -> None:
        """Cleans the set of independent nodes.

        This function is mostly there to handle cases that are not implemented.
        This function might remove nodes from the set of independent nodes.
        """
        assert self._independent_nodes is not None  # silence MyPy

        independent_nodes_were_updated = True
        while independent_nodes_were_updated:
            independent_nodes_were_updated = False
            for node in list(self._independent_nodes):
                # The only nodes that are important here, are the independent nodes
                #  at the boundaries, i.e. independent nodes that are connected to
                #  dependent nodes.
                if all(
                    oedge.dst in self._independent_nodes
                    for oedge in state.out_edges(node)
                    if not oedge.data.is_empty()
                ):
                    continue

                if isinstance(node, dace_nodes.AccessNode):
                    # This is actually the only case that is implemented, which is also the
                    #  only case that makes sense, as the AccessNode is needed as cache.
                    pass

                elif isinstance(node, dace_nodes.Tasklet):
                    # A Tasklet "generates" some data that has to be stored somewhere, for
                    #  example in an AccessNode, that has to be independent. Thus we will
                    #  now remove the node from the set of independent nodes.
                    self._independent_nodes.remove(node)
                    independent_nodes_were_updated = True
                    break

                else:
                    # We can not handle this kind of boundary node, so remove it.
                    self._independent_nodes.remove(node)
                    independent_nodes_were_updated = True
                    break

    def _rewire_map_scope(
        self,
        outer_entry: dace_nodes.MapEntry,
        outer_exit: dace_nodes.MapExit,
        inner_entry: dace_nodes.MapEntry,
        inner_exit: dace_nodes.MapExit,
        state: dace.SDFGState,
        sdfg: dace.SDFG,
    ) -> None:
        """Rewire the edges inside the scope defined by the outer map.

        The function assumes that the outer and inner map were obtained by a call
        to `_prepare_inner_outer_maps()`. The function will now rewire the connections of these
        nodes such that the dependent nodes are inside the scope of the inner map,
        while the independent nodes remain outside.

        Args:
            outer_entry: The entry node of the outer map.
            outer_exit: The exit node of the outer map.
            inner_entry: The entry node of the inner map.
            inner_exit: The exit node of the inner map.
            state: The state of the map.
            sdfg: The SDFG we operate on.
        """
        assert self._independent_nodes is not None and self._dependent_nodes is not None

        # Contains the nodes that are already have been handled.
        relocated_nodes: set[dace_nodes.Node] = set()

        # We now handle all independent nodes, this means that all of their
        #  _output_ edges have to go through the new inner map and the Memlets
        #  need modifications, because of the block parameter.
        for independent_node in self._independent_nodes:
            for out_edge in list(state.out_edges(independent_node)):
                edge_dst: dace_nodes.Node = out_edge.dst
                relocated_nodes.add(edge_dst)

                # If destination of this edge is also independent we do not need
                #  to handle it, because that node will also be before the new
                #  inner serial map.
                if edge_dst in self._independent_nodes:
                    continue

                # Special case, when we encounter an AccessNode, referring to a scalar,
                #  that is connected to the outer MapExit node, then we will insert
                #  a copy Tasklet, thus the copy Tasklet will then serve as dependent
                #  node that writes into the output.
                if isinstance(independent_node, dace_nodes.AccessNode) and edge_dst is outer_exit:
                    assert isinstance(independent_node.desc(sdfg), dace_data.Scalar)
                    assert not out_edge.data.is_empty()

                    copy_tlet = state.add_tasklet(
                        name=f"loop_blocking_copy_tlet_{independent_node.data}_{id(out_edge)}",
                        inputs={"__in"},
                        outputs={"__out"},
                        code="__out = __in",
                    )

                    # remove the current output edge from the state, but we need it.
                    org_out_edge = out_edge
                    state.remove_edge(org_out_edge)

                    # Now create a new `out_node`, that connects the independent node
                    #  with the copy Tasklet. We can use a plain Memlet for that.
                    #  This is the edge that we will split later.
                    out_edge = state.add_edge(
                        independent_node,
                        org_out_edge.src_conn,
                        copy_tlet,
                        "__in",
                        dace.Memlet(f"{independent_node.data}[0]"),
                    )

                    # Create the edge that connects the copy Tasklet with the outer MapExit.
                    #  Here we are copying the original Memlet, however, because the
                    #  new source is a Tasklet, instead of an AccessNode we must clear
                    #  the source subset.
                    new_output_edge = state.add_edge(
                        copy_tlet,
                        "__out",
                        outer_exit,
                        org_out_edge.dst_conn,
                        dace.Memlet.from_memlet(org_out_edge.data),
                    )
                    new_output_edge.data.src_subset = None
                    assert new_output_edge.data.dst_subset is not None

                    # Update `edge_dst` and mark the copy Tasklet as processed.
                    edge_dst = copy_tlet
                    relocated_nodes.add(copy_tlet)

                # Now split `out_edge` such that it passes through the new inner entry.
                #  We do not need to modify the subsets, i.e. replacing the variable
                #  on which we block, because the node is independent and the outgoing
                #  new inner map entry iterate over the blocked variable.
                if out_edge.data.is_empty():
                    # `out_edge` is an empty Memlet that ensures its source, which is
                    #  independent, is sequenced before its destination, which is
                    #  dependent. We now have to split the Memlet into two.
                    # TODO(phimuell): Can we remove this edge? Is the map enough to
                    #   ensure proper sequencing?
                    new_in_conn = None
                    new_out_conn = None
                    new_memlet_outside = dace.Memlet()

                elif not isinstance(independent_node, dace_nodes.AccessNode):
                    # For syntactical reasons the boundary, i.e. independent nodes
                    #  that have a connection to a dependent node, must be AccessNodes,
                    #  because something is needed as cache. Thus if you hit this
                    #  case then there is a bug in the `classify_node()` function
                    #  or your SDFG is wrong.
                    # NOTE: We do not allow direct connections between Tasklets.
                    raise NotImplementedError()

                else:
                    # NOTE: This creates more connections that are ultimately
                    #  necessary. However, figuring out which one to use and if
                    #  it would be valid, is very complicated, so we don't do it.
                    new_map_conn = inner_entry.next_connector(try_name=out_edge.data.data)
                    new_in_conn = "IN_" + new_map_conn
                    new_out_conn = "OUT_" + new_map_conn
                    new_memlet_outside = dace.Memlet.from_array(
                        out_edge.data.data, sdfg.arrays[out_edge.data.data]
                    )
                    inner_entry.add_in_connector(new_in_conn)
                    inner_entry.add_out_connector(new_out_conn)

                state.add_edge(
                    out_edge.src,
                    out_edge.src_conn,
                    inner_entry,
                    new_in_conn,
                    new_memlet_outside,
                )
                state.add_edge(
                    inner_entry,
                    new_out_conn,
                    out_edge.dst,
                    out_edge.dst_conn,
                    copy.deepcopy(out_edge.data),
                )
                state.remove_edge(out_edge)

        # Now we handle the dependent nodes, they differ from the independent nodes
        #  in that they _after_ the new inner map entry. Thus, we have to modify
        #  their incoming edges.
        for dependent_node in self._dependent_nodes:
            for in_edge in state.in_edges(dependent_node):
                edge_src: dace_nodes.Node = in_edge.src

                # The incoming edge of a dependent node (before any processing) either
                #  starts at:
                #   - The outer map.
                #   - An other dependent node.
                #   - An independent node.
                #  The last case was already handled by the loop above.
                if edge_src is inner_entry:
                    # Edge originated originally at an independent node, but was
                    #  already handled by the loop above.
                    assert dependent_node in relocated_nodes

                elif edge_src is not outer_entry:
                    # Edge originated at an other dependent node. There is nothing
                    #  that we have to do.
                    # NOTE: We can not test if `edge_src` is in `self._dependent_nodes`
                    #  because it only contains the dependent nodes that are directly
                    #  connected to the map entry.
                    assert edge_src not in self._independent_nodes

                elif in_edge.data.is_empty():
                    # The dependent node has an empty Memlet to the other map.
                    #  Since the inner map is sequenced after the outer map,
                    #  we will simply reconnect the edge to the inner map.
                    # TODO(phimuell): Are there situations where this makes problems.
                    dace_helpers.redirect_edge(state=state, edge=in_edge, new_src=inner_entry)

                elif edge_src is outer_entry:
                    # This dependent node originated at the outer map. Thus we have to
                    #  split the edge, such that it now passes through the inner map.
                    new_map_conn = inner_entry.next_connector(try_name=in_edge.src_conn[4:])
                    new_in_conn = "IN_" + new_map_conn
                    new_out_conn = "OUT_" + new_map_conn
                    new_memlet_inner = dace.Memlet.from_array(
                        in_edge.data.data, sdfg.arrays[in_edge.data.data]
                    )
                    state.add_edge(
                        in_edge.src,
                        in_edge.src_conn,
                        inner_entry,
                        new_in_conn,
                        new_memlet_inner,
                    )
                    state.add_edge(
                        inner_entry,
                        new_out_conn,
                        in_edge.dst,
                        in_edge.dst_conn,
                        copy.deepcopy(in_edge.data),
                    )
                    inner_entry.add_in_connector(new_in_conn)
                    inner_entry.add_out_connector(new_out_conn)
                    state.remove_edge(in_edge)

                else:
                    raise NotImplementedError("Unknown node configuration.")

        # In certain cases it might happen that we need to create an empty
        #  Memlet between the outer map entry and the inner one.
        if state.in_degree(inner_entry) == 0:
            state.add_edge(
                outer_entry,
                None,
                inner_entry,
                None,
                dace.Memlet(),
            )

        # Handle the Map exits
        #  This is simple reconnecting, there would be possibilities for improvements
        #  but we do not use them for now.
        for in_edge in state.in_edges(outer_exit):
            edge_conn = inner_exit.next_connector(in_edge.dst_conn[3:])
            dace_helpers.redirect_edge(
                state=state,
                edge=in_edge,
                new_dst=inner_exit,
                new_dst_conn="IN_" + edge_conn,
            )
            state.add_edge(
                inner_exit,
                "OUT_" + edge_conn,
                outer_exit,
                in_edge.dst_conn,
                copy.deepcopy(in_edge.data),
            )
            inner_exit.add_in_connector("IN_" + edge_conn)
            inner_exit.add_out_connector("OUT_" + edge_conn)

        # There is an invalid cache state in the SDFG, that makes the memlet
        #  propagation fail, to clear the cache we call the hash function.
        #  See: https://github.com/spcl/dace/issues/1703
        _ = sdfg.reset_cfg_list()
        dace_sdutils.canonicalize_memlet_trees_for_map(state=state, map_node=outer_entry)
        dace_propagation.propagate_memlets_map_scope(sdfg, state, outer_entry)

    def _check_if_blocking_is_favourable(
        self,
        state: dace.SDFGState,
    ) -> bool:
        """Test if the nodes are really independent nodes.

        After the classification the function will examine the set to see if some
        nodes were found that brings no benefit to move out. The classical example
        is a Tasklet that writes a constant into an AccessNode. These kind of
        nodes are filtered out.

        The function returns `True` if it decides that blocking is good and `False`
        otherwise. The function will not modify `self._independent_nodes`.
        """
        assert self._independent_nodes is not None
        assert self._dependent_nodes is None

        # There is nothing to move out so ignore it.
        if len(self._independent_nodes) == 0:
            return False

        # Currently we only filter out Tasklets that do not read any data, which
        #  is the example above, Because of how DaCe works we also subtract all
        #  of its output nodes, that are classified independent.
        # TODO(phimuell): Think if we should expand on that.
        nb_independent_nodes = len(self._independent_nodes)

        for node in self._independent_nodes:
            if isinstance(node, dace_nodes.Tasklet):
                if not all(iedge.data.is_empty() for iedge in state.in_edges(node)):
                    continue
                nb_independent_nodes -= 1
                for oedge in state.out_edges(node):
                    assert isinstance(oedge.dst, dace_nodes.AccessNode)
                    assert oedge.dst in self._independent_nodes
                    nb_independent_nodes -= 1
            assert nb_independent_nodes >= 0

        return nb_independent_nodes > 0



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause


from collections.abc import Generator
from contextlib import contextmanager
from typing import Mapping

from gt4py import eve
from gt4py.cartesian.gtc import common as gtc_common, definitions as gtc_definitions, oir
from gt4py.cartesian.gtc.passes.oir_optimizations import utils as optimization_utils
from gt4py.eve import codegen, concepts


def compute_extents(
    node: oir.Stencil,
) -> tuple[dict[str, gtc_definitions.Extent], dict[int, gtc_definitions.Extent]]:
    ctx: optimization_utils.StencilExtentComputer.Context = (
        optimization_utils.StencilExtentComputer().visit(node)
    )
    return ctx.fields, ctx.blocks


class DebugCodeGen(eve.VisitorWithSymbolTableTrait):
    def __init__(self) -> None:
        self.body = codegen.TextBlock()

    def visit_Stencil(
        self, stencil: oir.Stencil, symtable: Mapping[str, oir.FieldDecl], **_
    ) -> str:
        field_extents, block_extents = compute_extents(stencil)

        self.generate_imports()
        self.generate_run_function(stencil)
        self.initial_declarations(stencil, field_extents)

        self.generate_stencil_code(stencil, block_extents, symtable)

        return self.body.text

    def generate_imports(self) -> None:
        self.body.append("import numpy as np")
        self.body.empty_line()
        self.body.append("from gt4py.cartesian.gtc import ufuncs")
        self.body.append("from gt4py.cartesian.utils import Field")

    def generate_run_function(self, stencil: oir.Stencil) -> None:
        function_signature = "def run(*"
        args = []
        for param in stencil.params:
            args.append(self.visit(param))
        function_signature = ",".join([function_signature, *args])
        function_signature += ",_domain_, _origin_):"
        self.body.append(function_signature)
        self.body.indent()

    def initial_declarations(
        self, stencil: oir.Stencil, field_extents: dict[str, gtc_definitions.Extent]
    ) -> None:
        self.body.append("# ===== Domain Description ===== #")
        self.body.append("i_0, j_0, k_0 = 0, 0, 0")
        self.body.append("i_size, j_size, k_size = _domain_")
        self.body.empty_line()
        self.body.append("# ===== Temporary Declaration ===== #")
        self.generate_temp_decls(stencil.declarations, field_extents)
        self.body.empty_line()
        self.body.append("# ===== Field Declaration ===== #")
        self.generate_field_decls(stencil.params)
        self.body.empty_line()

    def generate_temp_decls(
        self,
        temporary_declarations: list[oir.Temporary],
        field_extents: dict[str, gtc_definitions.Extent],
    ) -> None:
        for declaration in temporary_declarations:
            self.body.append(self.visit(declaration, field_extents=field_extents))

    def generate_field_decls(self, declarations: list[oir.Decl]) -> None:
        for declaration in declarations:
            if isinstance(declaration, oir.FieldDecl):
                self.body.append(
                    f"{declaration.name} = Field({declaration.name}, _origin_['{declaration.name}'], "
                    f"({', '.join([str(x) for x in declaration.dimensions])}))"
                )

    # The visitors for all the control flow code that write directly to the `self.body` are in the following section
    def generate_stencil_code(
        self,
        stencil: oir.Stencil,
        block_extents: dict[int, gtc_definitions.Extent],
        symtable: Mapping[str, oir.FieldDecl],
    ) -> None:
        for loop in stencil.vertical_loops:
            for section in loop.sections:
                for execution in section.horizontal_executions:
                    with self.generate_ij_loop(block_extents, execution):
                        with self.create_k_loop_code(section, loop, symtable=symtable):
                            self.visit(execution, symtable=symtable)

    @contextmanager
    def create_k_loop_code(
        self,
        section: oir.VerticalLoopSection,
        loop: oir.VerticalLoop,
        symtable: Mapping[str, oir.FieldDecl],
    ) -> Generator:
        loop_bounds: str = self.visit(
            section.interval, var="k", direction=loop.loop_order, symtable=symtable
        )
        increment = "-1" if loop.loop_order == gtc_common.LoopOrder.BACKWARD else "1"
        loop_code = f"for k in range( {loop_bounds} , {increment}):"
        self.body.append(loop_code)
        self.body.indent()
        yield
        self.body.dedent()

    @contextmanager
    def generate_ij_loop(
        self, block_extents: dict[int, gtc_definitions.Extent], execution: oir.HorizontalExecution
    ) -> Generator:
        extents = block_extents[id(execution)]
        i_loop = f"for i in range(i_0 + {extents[0][0]} , i_size + {extents[0][1]}):"
        self.body.append(i_loop)
        self.body.indent()
        j_loop = f"for j in range(j_0 + {extents[1][0]} , j_size + {extents[1][1]}):"
        self.body.append(j_loop)
        self.body.indent()
        yield
        self.body.dedent()
        self.body.dedent()

    def visit_While(self, while_node: gtc_common.While, **kwargs) -> None:
        while_condition = self.visit(while_node.cond, **kwargs)
        while_code = f"while {while_condition}:"
        self.body.append(while_code)
        with self.body.indented():
            for statement in while_node.body:
                self.visit(statement, **kwargs)

    def visit_HorizontalExecution(
        self, horizontal_execution: oir.HorizontalExecution, **kwargs
    ) -> None:
        for statement in horizontal_execution.body:
            self.visit(statement, **kwargs)

    def visit_HorizontalMask(self, horizontal_mask: gtc_common.HorizontalMask, **kwargs) -> None:
        i_min, i_max = self.visit(horizontal_mask.i, var="i", **kwargs)
        j_min, j_max = self.visit(horizontal_mask.j, var="j", **kwargs)
        conditions = []
        if i_min is not None:
            conditions.append(f"({i_min}) <= i")
        if i_max is not None:
            conditions.append(f"i < ({i_max})")
        if j_min is not None:
            conditions.append(f"({j_min}) <= j")
        if j_max is not None:
            conditions.append(f"j < ({j_max})")
        assert len(conditions)
        if_code = f"if( {' and '.join(conditions)} ):"
        self.body.append(if_code)

    def visit_HorizontalRestriction(
        self, horizontal_restriction: oir.HorizontalRestriction, **kwargs
    ) -> None:
        self.visit(horizontal_restriction.mask, **kwargs)
        with self.body.indented():
            self.visit(horizontal_restriction.body, **kwargs)

    def visit_MaskStmt(self, mask_statement: oir.MaskStmt, **kwargs) -> None:
        self.body.append(f"if {self.visit(mask_statement.mask, **kwargs)}:")
        with self.body.indented():
            for statement in mask_statement.body:
                self.visit(statement, **kwargs)

    def visit_AssignStmt(self, assignment_statement: oir.AssignStmt, **kwargs) -> None:
        self.body.append(
            f"{self.visit(assignment_statement.left, **kwargs, is_write_access=True)} = {self.visit(assignment_statement.right, **kwargs)}"
        )

    # The visitors for the rest of the code-generation all return their strings directly and are in the following section
    def visit_VerticalLoop(self) -> None:
        pass

    def visit_FieldDecl(self, field_decl: oir.FieldDecl, **_) -> str:
        return str(field_decl.name)

    def visit_AxisBound(self, axis_bound: gtc_common.AxisBound, **kwargs) -> str:
        if axis_bound.level == gtc_common.LevelMarker.START:
            return f"{kwargs['var']}_0 + {axis_bound.offset}"
        if axis_bound.level == gtc_common.LevelMarker.END:
            return f"{kwargs['var']}_size + {axis_bound.offset}"

    def visit_RuntimeAxisBound(self, axis_bound: gtc_common.RuntimeAxisBound, **kwargs) -> str:
        return f"int({kwargs['var']}_0 + {self.visit(axis_bound.offset, **kwargs)})"

    def visit_Interval(self, interval: oir.Interval, **kwargs) -> str:
        if kwargs["direction"] == gtc_common.LoopOrder.BACKWARD:
            return ",".join(
                [
                    f"{self.visit(interval.end, **kwargs)} - 1",
                    f"{self.visit(interval.start, **kwargs)} - 1",
                ]
            )
        return ",".join([self.visit(interval.start, **kwargs), self.visit(interval.end, **kwargs)])

    def visit_Temporary(self, temporary_declaration: oir.Temporary, **kwargs) -> str:
        # Cartesian IJ
        field_extents = kwargs["field_extents"]
        local_field_extent = field_extents[temporary_declaration.name]
        i_padding: int = local_field_extent[0][1] - local_field_extent[0][0]
        j_padding: int = local_field_extent[1][1] - local_field_extent[1][0]
        shape: list[str] = [f"i_size + {i_padding}", f"j_size + {j_padding}"]
        field_offset = tuple(-ext[0] for ext in local_field_extent)
        offset = [str(off) for off in field_offset]

        # Cartesian vertical dimension K
        if temporary_declaration.dimensions[2]:
            shape.append("k_size")
            offset.append("0")

        # Data dimensions
        data_dimensions: list[str] = [str(dim) for dim in temporary_declaration.data_dims]
        offset += ["0"] * len(temporary_declaration.data_dims)

        # All together to write the `Field.empty` call
        dims = temporary_declaration.dimensions
        shape_decl = ", ".join(shape + data_dimensions)
        dtype: str = self.visit(temporary_declaration.dtype)

        return f"{temporary_declaration.name} = Field.empty(({shape_decl}), {dtype}, ({', '.join(offset)}), {dims})"

    def visit_DataType(self, data_type: gtc_common.DataType, **_) -> str:
        if data_type in {gtc_common.DataType.BOOL}:
            return data_type.name.lower()
        return f"np.{data_type.name.lower()}"

    def visit_VariableKOffset(self, variable_k_offset: oir.VariableKOffset, **kwargs) -> str:
        return f"i,j,k+int({self.visit(variable_k_offset.k, **kwargs)})"

    def visit_CartesianOffset(self, cartesian_offset: gtc_common.CartesianOffset, **kwargs) -> str:
        if "dimensions" in kwargs.keys():
            dimensions = kwargs["dimensions"]
            dimension_strings = []
            if dimensions[0]:
                dimension_strings.append(f"i + {cartesian_offset.i}")
            if dimensions[1]:
                dimension_strings.append(f"j + {cartesian_offset.j}")
            if dimensions[2]:
                dimension_strings.append(f"k + {cartesian_offset.k}")
            return ",".join(dimension_strings)
        return f"i + {cartesian_offset.i}, j + {cartesian_offset.j}, k + {cartesian_offset.k}"

    def visit_SymbolRef(self, symbol_ref: concepts.SymbolRef) -> str:
        return symbol_ref

    def visit_FieldAccess(
        self,
        field_access: gtc_common.FieldAccess,
        *,
        symtable: Mapping[str, oir.FieldDecl],
        **kwargs,
    ) -> str:
        is_write_access = kwargs.pop("is_write_access", False)

        if str(field_access.name) in symtable:
            dimensions = symtable[str(field_access.name)].dimensions
            kwargs.pop("dimensions", None)
            offset_str = self.visit(
                field_access.offset, dimensions=dimensions, symtable=symtable, **kwargs
            )
        else:
            offset_str = self.visit(field_access.offset, symtable=symtable, **kwargs)

        # Since the Field-class pads the shape of lower dimensional fields to always be 3d,
        # we need value extraction for the RHS otherwise we access arrays.
        # This leads to exceptions since numpy 2.0
        if not all(dimensions) and not is_write_access:
            item_extractor = ".item()"
        else:
            item_extractor = ""

        if field_access.data_index:
            data_index_access = ",".join(
                [
                    self.visit(data_index, symtable=symtable, **kwargs)
                    for data_index in field_access.data_index
                ]
            )
            if offset_str == "":
                return f"{field_access.name}[{data_index_access}]{item_extractor}"
            return f"{field_access.name}[{offset_str},{data_index_access}]{item_extractor}"
        return f"{field_access.name}[{offset_str}]{item_extractor}"

    def visit_AbsoluteKIndex(self, absolute_k_index: oir.AbsoluteKIndex, **kwargs) -> str:
        access_pattern = []
        if kwargs["dimensions"][2] is False:
            raise ValueError(
                "Tried accessing a field with no K-dimensions with an absolute K-index."
            )
        if kwargs["dimensions"][0]:
            access_pattern.append("i")
        if kwargs["dimensions"][1]:
            access_pattern.append("j")
        access_pattern.append(f"int({self.visit(absolute_k_index.k, **kwargs)})")
        return ",".join(access_pattern)

    def visit_BinaryOp(self, binary: oir.BinaryOp, **kwargs) -> str:
        return f"( {self.visit(binary.left, **kwargs)} {binary.op} {self.visit(binary.right, **kwargs)} )"

    def visit_Literal(self, literal: oir.Literal, **_) -> str:
        if literal.dtype == gtc_common.DataType.BOOL:
            literal_value = "True" if literal.value == gtc_common.BuiltInLiteral.TRUE else "False"
        else:
            literal_value = str(literal.value)

        if literal.dtype.bit_count() != 4:
            return f"{self.visit(literal.dtype)}({literal_value})"

        return literal_value

    def visit_Cast(self, cast: oir.Cast, **kwargs) -> str:
        return f"{self.visit(cast.dtype, **kwargs)}({self.visit(cast.expr, **kwargs)})"

    def visit_ScalarAccess(self, scalar_access: oir.ScalarAccess, **_) -> str:
        return scalar_access.name

    def visit_ScalarDecl(self, scalar_declaration: oir.ScalarDecl, **_) -> str:
        return scalar_declaration.name

    def visit_HorizontalInterval(
        self, horizontal_interval: gtc_common.HorizontalInterval, **kwargs
    ) -> tuple[str | None, str | None]:
        return self.visit(
            horizontal_interval.start, **kwargs
        ) if horizontal_interval.start else None, self.visit(
            horizontal_interval.end, **kwargs
        ) if horizontal_interval.end else None

    def visit_NativeFuncCall(self, native_function_call: oir.NativeFuncCall, **kwargs) -> str:
        arglist = [self.visit(arg, **kwargs) for arg in native_function_call.args]
        arguments = ",".join(arglist)
        function = gtc_common.OP_TO_UFUNC_NAME[gtc_common.NativeFunction][native_function_call.func]
        return f"ufuncs.{function}({arguments})"

    def visit_IteratorAccess(self, iterator_access: oir.IteratorAccess, **_) -> str:
        """Returns the axis in lower letter (to match the index of the loop)"""
        return iterator_access.name.value.lower()

    def visit_UnaryOp(self, unary_operator: oir.UnaryOp, **kwargs) -> str:
        return f"{unary_operator.op.value} {self.visit(unary_operator.expr, **kwargs)}"

    def visit_TernaryOp(self, ternary_operator: oir.TernaryOp, **kwargs) -> str:
        return f"{self.visit(ternary_operator.true_expr, **kwargs)} if {self.visit(ternary_operator.cond, **kwargs)} else {self.visit(ternary_operator.false_expr, **kwargs)}"

    # The visitors that are blacklisted for all the things that should not occur are here
    def visit_LocalScalar(self, local_scalar: oir.LocalScalar, **__) -> None:
        raise NotImplementedError(
            "This state should not be reached because LocalTemporariesToScalars should not have been called."
        )

    def visit_CacheDesc(self, cache_descriptor: oir.CacheDesc, **_) -> None:
        raise NotImplementedError("CacheDescriptors should never be visited in the debug backends")

    def visit_IJCache(self, ij_cache: oir.IJCache, **_) -> None:
        raise NotImplementedError("IJCaches should never be visited in the debug backend.")

    def visit_KCache(self, k_cache: oir.KCache, **_) -> None:
        raise NotImplementedError("KCaches should never be visited in the debug backend.")

    def visit_VerticalLoopSection(
        self, vertical_loop_section: oir.VerticalLoopSection, **_
    ) -> None:
        raise NotImplementedError("Vertical Loop section is not in the right place.")

    def visit_UnboundedInterval(self, unbounded_interval: oir.UnboundedInterval, **_) -> None:
        raise NotImplementedError("Unbounded Intervals are not supported in the debug backend.")



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""Contains type definitions and data utilities used during lowering to SDFG."""

from __future__ import annotations

import dataclasses
from typing import Final, TypeAlias

import dace
from dace import nodes as dace_nodes, subsets as dace_subsets

from gt4py.eve.extended_typing import MaybeNestedInTuple
from gt4py.next import common as gtx_common
from gt4py.next.iterator import builtins as gtir_builtins
from gt4py.next.program_processors.runners.dace import sdfg_args as gtx_dace_args
from gt4py.next.program_processors.runners.dace.lowering import gtir_dataflow, gtir_domain
from gt4py.next.type_system import type_specifications as ts


@dataclasses.dataclass(frozen=True)
class FieldopData:
    """
    Abstraction to represent data (scalars, arrays) during the lowering to SDFG.

    Attribute 'local_offset' must always be set for `FieldType` data with a local
    dimension generated from neighbors access in unstructured domain, and indicates
    the name of the offset provider used to generate the list of neighbor values.

    Args:
        dc_node: DaCe access node to the data storage.
        gt_type: GT4Py type definition, which includes the field domain information.
        origin: Tuple of start indices, in each dimension, for `FieldType` data.
            Pass an empty tuple for `ScalarType` data or zero-dimensional fields.
    """

    dc_node: dace_nodes.AccessNode
    gt_type: ts.FieldType | ts.ScalarType
    origin: tuple[dace.symbolic.SymbolicType, ...]

    def __post_init__(self) -> None:
        """Implements a sanity check on the constructed data type."""
        assert (
            len(self.origin) == 0
            if isinstance(self.gt_type, ts.ScalarType)
            else len(self.origin) == len(self.gt_type.dims)
        )

    def get_local_view(
        self, domain: gtir_domain.FieldopDomain, sdfg: dace.SDFG
    ) -> gtir_dataflow.IteratorExpr | gtir_dataflow.MemletExpr:
        """Helper method to access a field in local view, given the compute domain of a field operator."""
        if isinstance(self.gt_type, ts.ScalarType):
            assert isinstance(self.dc_node.desc(sdfg), dace.data.Scalar)
            return gtir_dataflow.MemletExpr(
                dc_node=self.dc_node,
                gt_field=ts.FieldType(dims=[], dtype=self.gt_type),
                subset=dace_subsets.Range.from_string("0"),
            )

        if isinstance(self.gt_type, ts.FieldType):
            it_indices: dict[gtx_common.Dimension, gtir_dataflow.DataExpr]
            if isinstance(self.dc_node.desc(sdfg), dace.data.Scalar):
                assert len(self.gt_type.dims) == 0  # zero-dimensional field
                it_indices = {}
            else:
                # The invariant below is ensured by calling `make_field()` to construct `FieldopData`.
                # The `make_field` constructor converts any local dimension, if present, to `ListType`
                # element type, while leaving the field domain with all global dimensions.
                assert all(dim.kind != gtx_common.DimensionKind.LOCAL for dim in self.gt_type.dims)
                domain_dims = [domain_range.dim for domain_range in domain]
                domain_indices = gtir_domain.get_element_subset(
                    domain_dims, origin=None
                ).min_element()
                it_indices = {
                    dim: gtir_dataflow.SymbolExpr(index, INDEX_DTYPE)
                    for dim, index in zip(domain_dims, domain_indices)
                }
            field_origin = [
                (dim, dace.symbolic.SymExpr(0) if self.origin is None else self.origin[i])
                for i, dim in enumerate(self.gt_type.dims)
            ]
            return gtir_dataflow.IteratorExpr(
                self.dc_node, self.gt_type.dtype, field_origin, it_indices
            )

        raise NotImplementedError(f"Node type {type(self.gt_type)} not supported.")

    def get_symbol_mapping(
        self, dataname: str, sdfg: dace.SDFG
    ) -> dict[str, dace.symbolic.SymExpr]:
        """
        Helper method to create the symbol mapping for array storage in a nested SDFG.

        Args:
            dataname: Name of the data container insiode the nested SDFG.
            sdfg: The parent SDFG where the `FieldopData` object lives.

        Returns:
            Mapping from symbols in nested SDFG to the corresponding symbolic values
            in the parent SDFG. This includes the range start and stop symbols (used
            to calculate the array shape as range 'stop - start') and the strides.
        """
        if isinstance(self.gt_type, ts.ScalarType):
            return {}
        outer_desc = self.dc_node.desc(sdfg)
        assert isinstance(outer_desc, dace.data.Array)
        # origin and size of the local dimension, in case of a field with `ListType` data,
        # are assumed to be compiled-time values (not symbolic), therefore the start and
        # stop range symbols of the inner field only extend over the global dimensions.
        if isinstance(self.gt_type.dtype, ts.ListType):
            local_dim = self.gt_type.dtype.offset_type
            assert local_dim is not None
            all_dims = gtx_common.order_dimensions([*self.gt_type.dims, local_dim])
            globals_size = [
                size
                for dim, size in zip(all_dims, outer_desc.shape, strict=True)
                if dim != local_dim
            ]
        else:
            all_dims = self.gt_type.dims
            globals_size = outer_desc.shape

        symbol_mapping: dict[str, dace.symbolic.SymbolicType] = {}
        for dim, origin, size in zip(self.gt_type.dims, self.origin, globals_size, strict=True):
            symbol_mapping |= {
                gtx_dace_args.range_start_symbol(dataname, dim).name: origin,
                gtx_dace_args.range_stop_symbol(dataname, dim).name: (origin + size),
            }
        for dim, stride in zip(all_dims, outer_desc.strides, strict=True):
            symbol_mapping |= {
                gtx_dace_args.field_stride_symbol(dataname, dim).name: stride,
            }
        return symbol_mapping


FieldopResult: TypeAlias = MaybeNestedInTuple[FieldopData | None]
"""Result of a field operator, can be either a field or a tuple fields.

For tuple of fields, any of the nested fields can be None, in case it is not used
and therefore does not need to be computed. The information whether a field needs
to be computed or not is the result of GTIR domain inference, and it is stored in
the GTIR node annex domain.
"""


@dataclasses.dataclass(frozen=True)
class SymbolicData:
    gt_type: ts.ScalarType
    value: dace.symbolic.SymbolicType


INDEX_DTYPE: Final[dace.typeclass] = getattr(dace, gtir_builtins.INTEGER_INDEX_BUILTIN)
"""Data type used for field indexing."""



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""An optimization to convert npir.LocalScalarDecl to npir.TemporaryDecl."""

from dataclasses import dataclass
from typing import Dict

from gt4py import eve
from gt4py.cartesian import utils
from gt4py.cartesian.gtc import common
from gt4py.cartesian.gtc.definitions import Extent
from gt4py.cartesian.gtc.numpy import npir


@dataclass
class Temporary:
    name: str
    dtype: common.DataType
    extent: Extent
    dimensions: tuple[bool, bool, bool]


def _all_local_scalars_are_unique_type(stencil: npir.Computation) -> bool:
    all_declarations = utils.flatten(
        stencil.walk_values().if_isinstance(npir.HorizontalBlock).getattr("declarations").to_list()
    )

    name_to_dtype: Dict[str, common.DataType] = {}
    for decl in all_declarations:
        if decl.name in name_to_dtype:
            if decl.dtype != name_to_dtype[decl.name]:
                return False
        else:
            name_to_dtype[decl.name] = decl.dtype

    return True


class ScalarsToTemporaries(eve.NodeTranslator):
    def visit_LocalScalarAccess(
        self, node: npir.LocalScalarAccess, *, temps_from_scalars: Dict[str, Temporary]
    ) -> npir.FieldSlice:
        return npir.FieldSlice(
            name=node.name,
            i_offset=0,
            j_offset=0,
            k_offset=0,
            dtype=temps_from_scalars[node.name].dtype,
        )

    def visit_HorizontalBlock(
        self, node: npir.HorizontalBlock, *, temps_from_scalars: Dict[str, Temporary]
    ) -> npir.HorizontalBlock:
        for decl in node.declarations:
            if decl.name not in temps_from_scalars:
                temps_from_scalars[decl.name] = Temporary(
                    name=decl.name,
                    dtype=decl.dtype,
                    extent=node.extent,
                    dimensions=(True, True, True),
                )
            else:
                temps_from_scalars[decl.name].extent |= node.extent

        body = self.visit(node.body, temps_from_scalars=temps_from_scalars)

        # Block without declarations
        return node.copy(update={"body": body, "declarations": []})

    def visit_Computation(self, node: npir.Computation) -> npir.Computation:
        if not _all_local_scalars_are_unique_type(node):
            raise TypeError(
                "Local scalars exist with different types and same name. "
                "The numpy backend currently assumes this is not the case."
            )

        temps_from_scalars: Dict[str, Temporary] = {}

        vertical_passes = self.visit(node.vertical_passes, temps_from_scalars=temps_from_scalars)

        scalar_temp_decls = [
            npir.TemporaryDecl(
                name=d.name,
                offset=(-d.extent[0][0], -d.extent[1][0]),
                padding=(d.extent[0][1] - d.extent[0][0], d.extent[1][1] - d.extent[1][0]),
                dimensions=d.dimensions,
                dtype=d.dtype,
            )
            for d in temps_from_scalars.values()
        ]

        return node.copy(
            update={
                "vertical_passes": vertical_passes,
                "temp_decls": node.temp_decls + scalar_temp_decls,
            }
        )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause
"""
This module contains a mypy plugin that can be used in downstream code to simplify type checking.

Usage:

    ```toml
    # pyproject.toml
    [tool.mypy]
    plugins = ['gt4py.next.type_system.mypy_plugin']
    ```

The goal of this plugin is to reduce the amount of false positives from mypy that arise from correct
usage of GT4Py. The following are examples for such false positives:

Dimensions are not fields:

    IDim = gtx.Dimension("IDim")
    gtx.Field[gtx.Dims[IDim], float]  # IDim is not a valid type

mixed precision math / different ways of describing the same dtype:

    a: gtx.Field[gtx.Dims[IDim], gtx.float64]
    a * 0.1  # no operator overload for types 'float' and 'float64'
    # same happens for int

Every false positive fixed in here should have a test in 'typing_tests/test_next.yaml'.
The documentation for how to write tests in that format is at https://github.com/typeddjango/pytest-mypy-plugins.

The documentation on mypy plugins is at https://mypy.readthedocs.io/en/latest/extending_mypy.html
"""

from __future__ import annotations

import typing


def iter_dim_names() -> typing.Iterator[str]:
    """Go through the four distinct place holders, then yield _AnyDim for everything after."""
    yield "_DimA"
    yield "_DimB"
    yield "_DimC"
    yield "_DimD"
    while True:
        yield "_AnyDim"


# if we can not import mypy we are not type checking with mypy, so we can skip all this
try:
    from mypy import plugin as mplugin, types

    DIM_MAP: dict[str, types.Type] = {}

    FLOAT_TYPES = ["builtins.float", "numpy.float32", "numpy.float64"]
    INT_TYPES = [
        "builtins.int",
        "numpy.int32",
        "numpy.int64",
        "numpy.integer",
        "numpy.signedinteger",
    ]

    def fixup_dims_type(ctx: mplugin.AnalyzeTypeContext) -> types.Type:
        """
        Overwrite Dims[...] to contain actual types.

        Example:

            CellDim = Dimension("CellDim")  # this is not a type
            a: Field[Dims[CellDim], gtx.float]  # we transform this to Field[Dims[_DimA]] where _DimA *is* a type

        The actual types are defined in 'gt4py.next.common', if typing.TYPE_CHECKING is true.
        """
        module_name = "gt4py.next.common"
        dims = iter_dim_names()
        args = []
        if ctx.type.args:
            # replacement 'Dims[OneDim, OtherDim]' -> 'Dims[_DimA, _DimB]' happens here
            for arg in ctx.type.args:
                argname = getattr(arg, "name", "unknown")
                if argname not in DIM_MAP:
                    DIM_MAP[argname] = ctx.api.analyze_type(
                        ctx.api.named_type(f"{module_name}.{next(dims)}", [])
                    )
                args.append(DIM_MAP[argname])
        else:
            # do not accidentally replace 'Dims' -> 'Dims[Any]' (the former matches any number of dims, the latter only one)
            args = [types.UnpackType(typ=types.AnyType(types.TypeOfAny.explicit))]
        result = ctx.api.analyze_type(ctx.api.named_type("gt4py.next.common.Dims", args))
        return result

    def fixup_dims_from_typealiases(ctx: mplugin.AnalyzeTypeContext) -> types.Type:
        """
        Catch 'Dimension' instances that made it through other replacements, this seems to happen in type aliases.

        Example:

            T = TypeVar("T", bound=(float, float64))
            CellDim = Dimension("CellDim")
            CellField: TypeAlias = Field[Dims[CellDim], T]

        If we have seen this dimension instance before, reuse the same dim type in the replacement, else use `_AnyDim`
        """
        result: types.Type | types.AnyType = types.AnyType(
            types.TypeOfAny.explicit
        )  # Fallback to Any if _AnyDim is not found
        try:
            if ctx.type.name not in DIM_MAP:
                DIM_MAP[ctx.type.name] = ctx.api.analyze_type(
                    ctx.api.named_type("gt4py.next.common._AnyDim", [])
                )
            result = DIM_MAP[ctx.type.name]
        except AssertionError:  # this probably happens when a dim type is analyzed in a context from where _AnyDim is unreachable
            pass
        return result

    def blur_float_precision(ctx: mplugin.AnalyzeTypeContext) -> types.Type:
        """Turn everything into 'builtins.float'."""
        # Note(ricoh): have tried to return numpy dtypes from here but ran into some error from mypy
        return ctx.api.named_type("builtins.float", [])

    def blur_int_precision(ctx: mplugin.AnalyzeTypeContext) -> types.Type:
        """Turn everything into 'builtins.int'"""
        return ctx.api.named_type("builtins.int", [])

    class TreatDimensionsAsTypes(mplugin.Plugin):
        def get_type_analyze_hook(
            self, fullname: str
        ) -> typing.Callable[[mplugin.AnalyzeTypeContext], types.Type] | None:
            """
            Decide whether to return a callback which can modify type hints before mypy analyzes them.

            If a callback is returned, it has to return a mypy-representation of a valid type.
            """
            # replace Dims args with actual types
            if fullname == "gt4py.next.common.Dims":
                return fixup_dims_type
            # replace stray 'Dimension' instances
            elif fullname.endswith("Dim") and not fullname == "gt4py.next.common._AnyDim":
                return fixup_dims_from_typealiases
            # treat all float precision types the same (GT4Py dsl will catch actual problems)
            elif fullname in FLOAT_TYPES:
                return blur_float_precision
            # treat all int precision types the same (GT4Py dsl will catch actual problems)
            elif fullname in INT_TYPES:
                return blur_int_precision
            return None

    def plugin(version: str) -> type[mplugin.Plugin]:
        """
        This is the entry point mypy looks for if this module was pointed to in config as a plugin.
        """
        return TreatDimensionsAsTypes

except ImportError:
    pass



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
import functools
from collections.abc import Mapping
from typing import Any


@dataclasses.dataclass(frozen=True, kw_only=True)
class SourceCodeSpec:
    """
    Basic settings for any source programming language.

    Formatting will happen through ``eve.codegen.format_source``.
    For available formatting options, check the options of the
    specific formatter used depending on ``.formatter_key``.
    """

    source_language: str
    file_extension: str
    formatter_key: str | None = None
    formatter_options: Mapping[str, Any] | None = None


@dataclasses.dataclass(frozen=True, kw_only=True)
class HeaderAndSourceCodeSpec(SourceCodeSpec):
    """Add a header file extension setting on top of the basic set."""

    header_extension: str


@dataclasses.dataclass(frozen=True, kw_only=True)
class PythonCodeSpec(SourceCodeSpec):
    """Settings for Python language."""

    source_language: str = "python"
    file_extension: str = "py"
    formatter_key: str = "python"


@dataclasses.dataclass(frozen=True, kw_only=True)
class SDFGCodeSpec(SourceCodeSpec):
    """Settings for SDFGs."""

    source_language: str = "SDFG"
    file_extension: str = "sdfg"


@dataclasses.dataclass(frozen=True, kw_only=True)
class CPPLikeCodeSpec(HeaderAndSourceCodeSpec):
    """Settings for C++-like language."""


@dataclasses.dataclass(frozen=True, kw_only=True)
class CPPCodeSpec(CPPLikeCodeSpec):
    """Settings for C++ language."""

    source_language: str = "CXX"
    file_extension: str = "cpp"
    header_extension: str = "hpp"
    formatter_key: str = "cpp"
    formatter_options: Mapping[str, Any] = dataclasses.field(
        default_factory=functools.partial(dict, style="LLVM")
    )


@dataclasses.dataclass(frozen=True, kw_only=True)
class CUDACodeSpec(CPPLikeCodeSpec):
    """Settings for CUDA language."""

    source_language: str = "CUDA"
    file_extension: str = "cu"
    header_extension: str = "cuh"
    formatter_key: str = "cpp"
    formatter_options: Mapping[str, Any] = dataclasses.field(
        default_factory=functools.partial(dict, style="LLVM")
    )


@dataclasses.dataclass(frozen=True, kw_only=True)
class HIPCodeSpec(CPPLikeCodeSpec):
    """Settings for HIP language."""

    source_language: str = "HIP"
    file_extension: str = "hip"
    header_extension: str = "h"
    formatter_key: str = "cpp"
    formatter_options: Mapping[str, Any] = dataclasses.field(
        default_factory=functools.partial(dict, style="LLVM")
    )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from .caches import (
    FillFlushToLocalKCaches,
    IJCacheDetection,
    KCacheDetection,
    PruneKCacheFills,
    PruneKCacheFlushes,
)
from .horizontal_execution_merging import HorizontalExecutionMerging, OnTheFlyMerging
from .inlining import MaskInlining
from .mask_stmt_merging import MaskStmtMerging
from .pruning import NoFieldAccessPruning, UnreachableStmtPruning
from .temporaries import LocalTemporariesToScalars, WriteBeforeReadTemporariesToScalars
from .vertical_loop_merging import AdjacentLoopMerging


__all__ = [
    "AdjacentLoopMerging",
    "FillFlushToLocalKCaches",
    "HorizontalExecutionMerging",
    "IJCacheDetection",
    "KCacheDetection",
    "LocalTemporariesToScalars",
    "MaskInlining",
    "MaskStmtMerging",
    "NoFieldAccessPruning",
    "OnTheFlyMerging",
    "PruneKCacheFills",
    "PruneKCacheFlushes",
    "UnreachableStmtPruning",
    "WriteBeforeReadTemporariesToScalars",
]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""Implementation of GTScript: an embedded DSL in Python for stencil computations.

Interface functions to define and compile GTScript definitions and empty symbol
definitions for the keywords of the DSL.
"""

import sys


sys.modules["__externals__"] = sys.modules["gt4py.cartesian.__externals__"]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any

from gt4py.cartesian import definitions as gt_definitions
from gt4py.cartesian.frontend import gtscript_frontend, nodes


class GTScriptSyntaxError(gt_definitions.GTSyntaxError):
    def __init__(self, message: str | None, *, loc: nodes.Location | None = None) -> None:
        if not message:
            message = "Syntax error"
            if loc is not None:
                message = f"{message} in '{loc.scope}' (line: {loc.line}, col: {loc.column})"
        super().__init__(message, frontend=gtscript_frontend.GTScriptFrontend.name)
        self.loc = loc


class GTScriptSymbolError(GTScriptSyntaxError):
    def __init__(
        self, name: str, message: str | None = None, *, loc: nodes.Location | None = None
    ) -> None:
        if not message:
            message = f"Unknown symbol '{name}'"
            if loc is not None:
                message = f"{message} in '{loc.scope}' (line: {loc.line}, col: {loc.column})"
        super().__init__(message, loc=loc)
        self.name = name


class GTScriptDefinitionError(GTScriptSyntaxError):
    def __init__(
        self,
        name: str,
        value: Any,
        message: str | None = None,
        *,
        loc: nodes.Location | None = None,
    ) -> None:
        if not message:
            message = f"Invalid definition for '{name}' symbol"
            if loc is not None:
                message = f"{message} in '{loc.scope}' (line: {loc.line}, col: {loc.column})"
        super().__init__(message, loc=loc)
        self.name = name
        self.value = value


class GTScriptValueError(GTScriptDefinitionError):
    def __init__(
        self,
        name: str,
        value: Any,
        message: str | None = None,
        *,
        loc: nodes.Location | None = None,
    ) -> None:
        if not message:
            message = f"Invalid value for '{name}'"
            if loc is not None:
                message = f"{message} in '{loc.scope}' (line: {loc.line}, col: {loc.column})"
        super().__init__(name, value, message, loc=loc)


class GTScriptDataTypeError(GTScriptSyntaxError):
    def __init__(
        self,
        name: str,
        data_type: Any,
        message: str | None = None,
        *,
        loc: nodes.Location | None = None,
    ) -> None:
        if not message:
            message = f"Invalid data type for '{name}' numeric symbol "
            if loc is not None:
                message = f"{message} in '{loc.scope}' (line: {loc.line}, col: {loc.column})"
        super().__init__(message, loc=loc)
        self.name = name
        self.data_type = data_type


class GTScriptAssertionError(gt_definitions.GTSpecificationError):
    def __init__(self, source: list[str], *, loc: nodes.Location | None = None) -> None:
        if loc:
            message = f"Assertion failed at line {loc.line}, col {loc.column}:\n{source}"
        else:
            message = f"Assertion failed.\n{source}"
        super().__init__(message)
        self.loc = loc



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Iterator, Sequence

from gt4py.next import common, errors
from gt4py.next.ffront import (
    program_ast as past,
    stages as ffront_stages,
    type_specifications as ts_ffront,
)
from gt4py.next.otf import arguments, toolchain, workflow
from gt4py.next.type_system import type_info, type_specifications as ts


def transform_program_args(
    inp: ffront_stages.ConcretePASTProgramDef,
) -> ffront_stages.ConcretePASTProgramDef:
    rewritten_args, rewritten_kwargs = _process_args(
        past_node=inp.data.past_node, args=inp.args.args, kwargs=inp.args.kwargs
    )
    return toolchain.ConcreteArtifact(
        data=inp.data,
        args=arguments.CompileTimeArgs(
            args=rewritten_args,
            kwargs=rewritten_kwargs,
            offset_provider=inp.args.offset_provider,
            column_axis=inp.args.column_axis,
            argument_descriptor_contexts=inp.args.argument_descriptor_contexts,
        ),
    )


def transform_program_args_factory(
    cached: bool = True,
) -> workflow.Workflow[ffront_stages.ConcretePASTProgramDef, ffront_stages.ConcretePASTProgramDef]:
    wf = transform_program_args
    if cached:
        wf = workflow.CachedStep(wf, hash_function=ffront_stages.fingerprint_stage)
    return wf


def _validate_args(
    past_node: past.Program, arg_types: Sequence[ts.TypeSpec], kwarg_types: dict[str, ts.TypeSpec]
) -> None:
    if not isinstance(past_node.type, ts_ffront.ProgramType):
        raise TypeError("Can not validate arguments for PAST programs prior to type inference.")

    try:
        type_info.accepts_args(
            past_node.type, with_args=arg_types, with_kwargs=kwarg_types, raise_exception=True
        )
    except ValueError as err:
        raise errors.DSLError(
            None, f"Invalid argument types in call to '{past_node.id}'.\n{err}"
        ) from err


def _process_args(
    past_node: past.Program,
    args: Sequence[ts.TypeSpec],
    kwargs: dict[str, ts.TypeSpec],
) -> tuple[tuple[ts.TypeSpec], dict[str, ts.TypeSpec]]:
    if not isinstance(past_node.type, ts_ffront.ProgramType):
        raise TypeError("Can not process arguments for PAST programs prior to type inference.")

    args, kwargs = type_info.canonicalize_arguments(past_node.type, args, kwargs)
    _validate_args(past_node=past_node, arg_types=args, kwarg_types=kwargs)

    return args, kwargs


def _field_constituents_range_and_dims(
    arg: Any,  # TODO(havogt): improve typing
    arg_type: ts.DataType,
) -> Iterator[tuple[tuple[tuple[int, int], ...], list[common.Dimension]]]:
    match arg_type:
        case ts.TupleType():
            for el, el_type in zip(arg, arg_type.types):
                assert isinstance(el_type, ts.DataType)
                yield from _field_constituents_range_and_dims(el, el_type)
        case ts.FieldType():
            dims = type_info.extract_dims(arg_type)
            if isinstance(arg, ts.TypeSpec):  # TODO
                yield (tuple(), dims)
            elif dims:
                assert (
                    hasattr(arg, "domain")
                    and isinstance(arg.domain, common.Domain)
                    and len(arg.domain.dims) == len(dims)
                )
                yield (tuple((r.start, r.stop) for r in arg.domain.ranges), dims)
            else:
                yield from []  # ignore 0-dim fields
        case ts.ScalarType():
            yield from []  # ignore scalars
        case _:
            raise ValueError("Expected 'FieldType' or 'TupleType' thereof.")



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from .gtir_k_boundary import compute_k_boundary, compute_min_k_size
from .gtir_pipeline import GtirPipeline
from .gtir_prune_unused_parameters import prune_unused_parameters
from .horizontal_masks import compute_relative_mask, mask_overlap_with_extent
from .oir_access_kinds import compute_access_kinds
from .oir_pipeline import DefaultPipeline, OirPipeline


__all__ = [
    "DefaultPipeline",
    "GtirPipeline",
    "OirPipeline",
    "compute_access_kinds",
    "compute_k_boundary",
    "compute_min_k_size",
    "compute_relative_mask",
    "mask_overlap_with_extent",
    "prune_unused_parameters",
]
